# constants for exporter objects.  At this point there
# is a only one exporter type which depends on ogr. As ogr
# is an optional install the constants have been moved to a
# module of their own


try:
    from osgeo import ogr
except ImportError:
    ogr = None

if ogr is not None:
    OGR_FIELD_TYPE_MAP = {
        "int": ogr.OFTInteger,
        int: ogr.OFTInteger,
        "str": ogr.OFTString,
        str: ogr.OFTString,
        "real": ogr.OFTReal,
        "float": ogr.OFTReal,
        float: ogr.OFTReal,
        bool: ogr.OFTInteger,
    }

    OGR_GEOM_TYPE_MAP = {
        "point": ogr.wkbPoint,
        "line": ogr.wkbLineString,
        "multiline": ogr.wkbLineString,
        "bbox": ogr.wkbPolygon,
    }

SHP_DRIVER_NAME = "ESRI Shapefile"
GEO_PACKAGE_DRIVER_NAME = "GPKG"
GEOJSON_DRIVER_NAME = "GeoJSON"


CHANNELS_EXPORT_FIELDS = [
    "id",
    "content_pk",
    "code",
    "kcu",
    "calculation_type",
    "dist_calc_points",
    "connection_node_start_pk",
    "connection_node_end_pk",
    "discharge_coefficient",
]

PIPES_EXPORT_FIELDS = [
    "id",
    "content_pk",
    "display_name",
    "material",
    "kcu",
    "friction_value",
    "calculation_type",
    "sewerage_type",
    "cross_section_shape",
    "cross_section_width",
    "cross_section_height",
    "invert_level_start_point",
    "invert_level_end_point",
    "connection_node_start_pk",
    "connection_node_end_pk",
    "discharge_coefficient",
]

WEIRS_EXPORT_FIELDS = [
    "id",
    "content_pk",
    "display_name",
    "code",
    "kcu",
    "cross_section_shape",
    "cross_section_width",
    "cross_section_height",
    "crest_level",
    "crest_type",
    "sewerage",
    "discharge_coefficient_positive",
    "discharge_coefficient_negative",
    "friction_value",
    "friction_type",
    "connection_node_start_pk",
    "connection_node_end_pk",
]

CULVERT_EXPORT_FIELDS = [
    "id",
    "content_pk",
    "display_name",
    "code",
    "kcu",
    "cross_section_shape",
    "cross_section_width",
    "cross_section_height",
    "calculation_type",
    "friction_value",
    "friction_type",
    "dist_calc_points",
    "discharge_coefficient_positive",
    "discharge_coefficient_negative",
    "invert_level_start_point",
    "invert_level_end_point",
    "connection_node_start_pk",
    "connection_node_end_pk",
]

ORIFICES_EXPORT_FIELDS = [
    "id",
    "content_pk",
    "display_name",
    "kcu",
    "crest_level",
    "crest_type",
    "sewerage",
    "discharge_coefficient_negative",
    "discharge_coefficient_positive",
    "friction_type",
    "friction_value",
    "connection_node_start_pk",
    "connection_node_end_pk",
]

CONNECTION_NODES_EXPORT_FIELDS = [
    "content_pk",
    "initial_waterlevel",  # missing in prepare step
]

MANHOLE_EXPORT_FIELDS = [
    "id",
    "content_pk",
    "display_name",
    "calculation_type",
    "sumax",
    "bottom_level",
    "surface_level",
    "drain_level",
    "width",
    "shape",
    "storage_area",
    "initial_waterlevel",
]

PUMPS_EXPORT_FIELDS = [
    "id",
    "content_pk",
    "display_name",
    "start_level",
    "bottom_level",
    "lower_stop_level",
    "capacity",
    "type",
]

LEVEES_EXPORT_FIELDS = [
    "id",
    "crest_level",
    "max_breach_depth",
]

BREACHES_EXPORT_FIELDS = [
    "id",
    "code",
    "display_name",
    "content_pk",
    "levmat",
    "levbr",
    "kcu",
    "levl",
    "coordinates",
]

CELLS_EXPORT_FIELDS = [
    "id",
    "content_pk",
    "node_type",
    "sumax",
    "z_coordinate",
]

DEFAULT_EXPORT_FIELDS = {
    "Lines": "ALL",
    "Pipes": PIPES_EXPORT_FIELDS,
    "Channels": CHANNELS_EXPORT_FIELDS,
    "Weirs": WEIRS_EXPORT_FIELDS,
    "Culverts": CULVERT_EXPORT_FIELDS,
    "Orifices": ORIFICES_EXPORT_FIELDS,
    "Nodes": "ALL",
    "ConnectionNodes": CONNECTION_NODES_EXPORT_FIELDS,
    "Manholes": MANHOLE_EXPORT_FIELDS,
    "Cells": CELLS_EXPORT_FIELDS,
    "Grid": "ALL",
    "Breaches": BREACHES_EXPORT_FIELDS,
    "Levees": LEVEES_EXPORT_FIELDS,
    "Pumps": PUMPS_EXPORT_FIELDS,
}



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

import logging
import os

import numpy as np

try:
    from osgeo import ogr
except ImportError:
    ogr = None


from threedigrid.admin import exporter_constants as const
from threedigrid.geo_utils import get_spatial_reference
from threedigrid.orm.base.exporters import BaseOgrExporter

logger = logging.getLogger(__name__)


def get_geometry(field_name, field_type, data, index, **kwargs):
    geom = ogr.Geometry(const.OGR_GEOM_TYPE_MAP[field_type])

    if "__" in field_name:
        data[field_name] = kwargs.get(field_name, np.array([]))

    if field_type == "point":
        geom.AddPoint(data[field_name][0][index], data[field_name][1][index])
    elif field_type == "bbox":
        ring = ogr.Geometry(ogr.wkbLinearRing)
        ring.AddPoint(data[field_name][0][index], data[field_name][1][index])
        ring.AddPoint(data[field_name][2][index], data[field_name][1][index])

        ring.AddPoint(data[field_name][2][index], data[field_name][3][index])

        ring.AddPoint(data[field_name][0][index], data[field_name][3][index])

        ring.AddPoint(data[field_name][0][index], data[field_name][1][index])

        # Create polygon from ring
        geom.AddGeometry(ring)
    elif field_type == "line":
        view = data[field_name].astype("float64")
        # print(data[field_name][:, index], data[field_name][:, index].dtype)
        geom.AddPoint(view[0][index], view[1][index])
        geom.AddPoint(view[2][index], view[3][index])
    elif field_type == "multiline":
        for x, y in data[field_name][index].reshape(2, -1).T:
            geom.AddPoint_2D(x, y)
    else:
        raise Exception("Unknown field_type %s", field_type)
    return geom


def get_field_type(model, field_name):
    if "__" in field_name:
        field_name, _ = field_name.split("__")
    return model._get_field(field_name).type


class GpkgExporter(BaseOgrExporter):
    def __init__(self, model):
        """
        :param model: model instance (filtered or not)
        """
        self.model = model
        self.supported_drivers = {
            const.GEO_PACKAGE_DRIVER_NAME,
        }
        self.set_driver(const.GEO_PACKAGE_DRIVER_NAME)

    def add_meta_data(self, file_name, layer_name):
        """
        Note: assumes self.model to be gridadmin root object.
        """
        meta_data = dict(self.model.h5py_file.attrs.items())

        if os.path.exists(file_name):
            data_source = self.driver.Open(file_name, update=1)
        else:
            data_source = self.driver.CreateDataSource(file_name)

        layer = data_source.CreateLayer(
            layer_name, geom_type=ogr.wkbNone, options=["OVERWRITE=YES"]
        )

        usable_fields = []
        for key in meta_data.keys():
            field_type = type(meta_data[key].tolist())
            value = meta_data[key]

            if field_type == bytes:
                field_type = str
            if field_type == list:
                field_type = str
                value = ",".join([str(x) for x in value])

            if field_type in const.OGR_FIELD_TYPE_MAP:
                usable_fields.append((key, field_type, value))
                field = ogr.FieldDefn(key, const.OGR_FIELD_TYPE_MAP[field_type])
                if field_type in ("str", str):
                    field.SetWidth(2048)
                elif field_type in ("bool", bool):
                    field.SetSubType(ogr.OFSTBoolean)
                layer.CreateField(field)

        _definition = layer.GetLayerDefn()
        data_source.StartTransaction()

        feature = ogr.Feature(_definition)

        for key, field_type, value in usable_fields:
            self.set_field(feature, key, field_type, value)

        layer.CreateFeature(feature)
        data_source.CommitTransaction()
        data_source = None

    def save(
        self, file_name, layer_name, field_definitions, progress_func=None, **kwargs
    ):
        """
        save to file format specified by the driver, e.g. shapefile

        :param file_name: name of the outputfile

        :param kwargs: allows to override what's is used as values, see default.py for an example
        """
        field_map = field_definitions
        data = self.model.data

        total = data["id"].size

        if progress_func:
            progress_func(0, total)

        if self.driver is None:
            self.set_driver(extension=os.path.splitext(file_name)[1])

        assert self.driver is not None
        geomtype = 0

        # Pick reproject_to_epsg or original model epsg_code
        target_epsg_code = self.model.reproject_to_epsg or self.model.epsg_code

        sr = get_spatial_reference(target_epsg_code)

        if os.path.exists(file_name):
            data_source = self.driver.Open(file_name, update=1)
        else:
            data_source = self.driver.CreateDataSource(file_name)

        layer = data_source.CreateLayer(layer_name, sr, geomtype, ["OVERWRITE=YES"])

        for field_name, ogr_field_name in field_map.items():
            if ogr_field_name == "the_geom":
                continue

            field_type = get_field_type(self.model, field_name)

            if field_type in const.OGR_FIELD_TYPE_MAP:
                field = ogr.FieldDefn(
                    ogr_field_name, const.OGR_FIELD_TYPE_MAP[field_type]
                )
                if field_type in ("bool", bool):
                    field.SetSubType(ogr.OFSTBoolean)

                layer.CreateField(field)

            else:
                raise Exception(
                    "Could not find type for %s with type %s", field_name, field_type
                )

        _definition = layer.GetLayerDefn()

        data_source.StartTransaction()

        geom_def = [(x, field_map[x]) for x in field_map if field_map[x] == "the_geom"]

        if geom_def:
            geom_def = geom_def[0]
        else:
            geom_def = None

        for i in range(total):
            if data["id"][i] == 0:
                continue  # skip the dummy element

            feature = ogr.Feature(_definition)

            if geom_def:
                field_name, ogr_field_name = geom_def
                field_type = get_field_type(self.model, field_name)
                geom = get_geometry(field_name, field_type, data, i, **kwargs)
                feature.SetGeometry(geom)

            for field_name, ogr_field_name in field_map.items():
                field_type = get_field_type(self.model, field_name)

                if ogr_field_name == "the_geom":
                    continue

                if "__" in field_name:
                    field_name_split, attribute = field_name.split("__")

                    try:
                        raw_value = data[field_name_split][:, i][int(attribute)]
                    except (IndexError, ValueError):
                        raw_value = kwargs.get(field_name, None)
                else:
                    try:
                        raw_value = data[field_name][i]
                    except IndexError:
                        # Try kwargs else None
                        raw_value = kwargs.get(field_name, None)
                self.set_field(feature, ogr_field_name, field_type, raw_value)

            layer.CreateFeature(feature)
            feature = None
            if progress_func:
                progress_func(i + 1, total)

        data_source.CommitTransaction()
        data_source = None



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
"""
Models
++++++
In the threedicore lines are called flow lines and represent links between
calculation points. They can exist between 1d and 2d points, between surface
and groundwater nodes etc. The line type is described by the kcu attribute
For an overview of the kcu types see :ref:`kcu-label`.
"""


import numpy as np

from threedigrid.admin.lines import exporters, subsets
from threedigrid.orm.base.fields import IndexArrayField
from threedigrid.orm.fields import ArrayField, LineArrayField, MultiLineArrayField
from threedigrid.orm.models import Model

LINE_SUBSETS = {"kcu__in": subsets.KCU__IN_SUBSETS}


def convert_pixel_coords(i, j, geo_transform):
    a, b, c, d, e, f = geo_transform
    x = a * i + b * j + c
    y = d * i + e * j + f
    return (x, y)


class Lines(Model):
    """
    fields originating from threedicore:

        - kcu (line type)
        - lik ()
        - line (index of node representing start/end point)

    added fields from spatialite database:

        - content_pk (primary key database)
        - content_type (type of object as stored in
          database, e.g v2_channel, v2_weir etc)

    custom fields
        - line_coords (coordinate pairs start/end point)
        - line_geometries (geometries from content_type)

    """

    kcu = ArrayField(type=int)
    lik = ArrayField(type=int)
    line = IndexArrayField(to="Nodes")
    dpumax = ArrayField(type=float)
    flod = ArrayField(type=float)
    flou = ArrayField(type=float)
    cross1 = IndexArrayField(to="CrossSections")
    cross2 = IndexArrayField(to="CrossSections")
    ds1d = ArrayField(type=float)
    ds1d_half = ArrayField(type=float)
    cross_weight = ArrayField(type=float)
    invert_level_start_point = ArrayField(type=float)
    invert_level_end_point = ArrayField(type=float)
    content_pk = ArrayField(type=int)
    content_type = ArrayField(type=str)
    zoom_category = ArrayField(type=int)
    cross_pix_coords = LineArrayField()
    line_coords = LineArrayField()
    line_geometries = MultiLineArrayField()
    discharge_coefficient_negative = ArrayField(type=float)
    discharge_coefficient_positive = ArrayField(type=float)
    sewerage = ArrayField(type=bool)
    sewerage_type = ArrayField(type=int)

    SUBSETS = LINE_SUBSETS

    GPKG_DEFAULT_FIELD_MAP = {
        "id": "id",
        "discharge_coefficient_positive": "discharge_coefficient_positive",
        "discharge_coefficient_negative": "discharge_coefficient_negative",
        "kcu": "line_type",
        "content_type": "source_table",
        "content_pk": "source_table_id",
        "invert_level_start_point": "invert_level_start_point",
        "invert_level_end_point": "invert_level_end_point",
        "dpumax": "exchange_level",
        "line__0": "calculation_node_id_start",
        "line__1": "calculation_node_id_end",
        "line_geometries": "the_geom",
        "sewerage": "sewerage",
        "sewerage_type": "sewerage_type",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._exporters = [
            exporters.LinesOgrExporter(self),
        ]

    def cross_pix_coords_transformed(self, geo_transform):
        """
        Get pix_coords transformed by geo_transform (=ga.grid.transform)

        >>> x = a * i + b * j + c
        >>> y = d * i + e * j + f

        """
        x, y = convert_pixel_coords(
            self.cross_pix_coords[0], self.cross_pix_coords[1], geo_transform
        )
        x2, y2 = convert_pixel_coords(
            self.cross_pix_coords[2], self.cross_pix_coords[3], geo_transform
        )
        return np.vstack([x, y, x2, y2])

    @property
    def pipes(self):
        return self._filter_as(Pipes, content_type="v2_pipe")

    @property
    def weirs(self):
        return self._filter_as(Weirs, content_type="v2_weir")

    @property
    def orifices(self):
        return self._filter_as(Orifices, content_type="v2_orifice")

    @property
    def channels(self):
        return self._filter_as(Channels, content_type="v2_channel")

    @property
    def culverts(self):
        return self._filter_as(Culverts, content_type="v2_culvert")

    @property
    def line_nodes(self):
        return np.dstack(self.line)[0]


class Pipes(Lines):
    display_name = ArrayField(type=str)
    friction_type = ArrayField(type=int)
    friction_value = ArrayField(type=float)
    material = ArrayField(type=int)
    cross_section_height = ArrayField(type=float)
    cross_section_width = ArrayField(type=float)
    cross_section_shape = ArrayField(type=float)
    sewerage_type = ArrayField(type=int)
    calculation_type = ArrayField(type=int)
    connection_node_start_pk = ArrayField(type=int)
    connection_node_end_pk = ArrayField(type=int)
    discharge_coefficient = ArrayField(type=float)


class Channels(Lines):
    code = ArrayField(type=str)
    calculation_type = ArrayField(type=int)
    dist_calc_points = ArrayField(type=float)
    connection_node_start_pk = ArrayField(type=int)
    connection_node_end_pk = ArrayField(type=int)
    discharge_coefficient = ArrayField(type=float)


class Weirs(Lines):
    code = ArrayField(type=str)
    display_name = ArrayField(type=str)
    sewerage = ArrayField(type=int)
    friction_type = ArrayField(type=int)
    friction_value = ArrayField(type=float)
    crest_type = ArrayField(type=int)
    crest_level = ArrayField(type=float)
    connection_node_start_pk = ArrayField(type=int)
    connection_node_end_pk = ArrayField(type=int)
    cross_section_height = ArrayField(type=float)
    cross_section_width = ArrayField(type=float)
    cross_section_shape = ArrayField(type=float)

    @property
    def line_coord_angles(self):
        # Filter raw values with content_type and content_pk
        # to always get the raw (metric) values
        # and not the reprojected ones to calculate
        # the angles with.
        content_type = self._datasource["content_type"][:]
        line_content_pk = self._datasource["content_pk"][:]
        pk_mask = np.isin(line_content_pk, self.content_pk)
        raw_values = self._datasource["line_coords"][:][
            :, (content_type == b"v2_weir") & pk_mask
        ]

        return self._get_field("line_coords").get_angles_in_degrees(raw_values)


class Culverts(Lines):
    display_name = ArrayField(type=str)
    code = ArrayField(type=str)
    calculation_type = ArrayField(type=int)
    dist_calc_points = ArrayField(type=float)
    cross_section_height = ArrayField(type=float)
    cross_section_width = ArrayField(type=float)
    cross_section_shape = ArrayField(type=float)
    friction_type = ArrayField(type=int)
    friction_value = ArrayField(type=float)
    connection_node_start_pk = ArrayField(type=int)
    connection_node_end_pk = ArrayField(type=int)


class Orifices(Lines):
    display_name = ArrayField(type=str)
    sewerage = ArrayField(type=int)
    friction_type = ArrayField(type=int)
    friction_value = ArrayField(type=float)
    crest_type = ArrayField(type=int)
    crest_level = ArrayField(type=float)
    connection_node_start_pk = ArrayField(type=int)
    connection_node_end_pk = ArrayField(type=int)



import unittest

import numpy as np
from shapely.geometry import LineString, Point, Polygon

from threedigrid.orm.base.fields import ArrayField
from threedigrid.orm.base.filters import (
    BaseCompareFilter,
    BaseFilter,
    EqualsFilter,
    get_filter,
    InFilter,
)
from threedigrid.orm.fields import (
    BboxArrayField,
    LineArrayField,
    MultiLineArrayField,
    PointArrayField,
    PolygonArrayField,
)
from threedigrid.orm.filters import FILTER_MAP, GeometryIntersectionFilter, PointFilter


class BasicFilterTests(unittest.TestCase):
    def setUp(self):
        self.field = ArrayField()

    def test_base_filter(self):
        b = BaseFilter()
        with self.assertRaises(NotImplementedError):
            b.filter("foo")

    def test_repr(self):
        f = BaseCompareFilter("in", self.field, [1, 2, 3])
        self.assertTrue(repr(f))

    def test_get_filter_eq(self):
        """Test that we get a default EqualsFilter without keys."""
        f = get_filter(["content_type"], self.field, 3)
        self.assertIsInstance(f, EqualsFilter)

        # also works when 'eq' is explicitly given
        f = get_filter(["content_type", "eq"], self.field, 3)
        self.assertIsInstance(f, EqualsFilter)

    def test_eq_filter(self):
        f = EqualsFilter("content_type", self.field, np.array([1, 2]))
        filtered = f.filter({"content_type": np.array([1, 2])})
        self.assertTrue(filtered.all())

        filtered = f.filter({"content_type": np.array([1, 3])})
        self.assertFalse(filtered.all())

    def test_get_filter_in(self):
        f = get_filter(["content_pk", "in"], self.field, [2])
        self.assertIsInstance(f, InFilter)

    def test_get_filter_error(self):
        """get_filter fails with malformed splitted keys."""
        with self.assertRaises(ValueError):
            get_filter(["content_pk", "in", "foo"], self.field, [2])

        with self.assertRaises(ValueError):
            get_filter([], self.field, [2])

    def test_get_nonexisting_filter(self):
        with self.assertRaises(ValueError):
            get_filter(["content_type", "not_a_filter"], self.field, 3)


class PointFilterTests(unittest.TestCase):
    def setUp(self):
        self.field = PolygonArrayField()

    def test_get_point_filter(self):
        f = get_filter(
            ["cell_coords", "contains_point"],
            self.field,
            [1, 5],
            filter_map=FILTER_MAP,
        )
        self.assertIsInstance(f, PointFilter)

    def test_point_filter(self):
        f = PointFilter("coord_cells", self.field, np.array([1, 5]))
        filtered = f.filter(
            {"coord_cells": np.array([[0, 2], [0, 2], [6, 6], [10, 11]])}
        )
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])


class GeometryIntersectionFilterTest(unittest.TestCase):
    def setUp(self):
        self.geometry = Polygon([[0, 0], [0, 1], [1, 1], [0, 0]])
        self.field = BboxArrayField()

    def test_get_intersect_geometry_filter(self):
        f = get_filter(
            ["cell_coords", "intersects_geometry"],
            BboxArrayField(),
            self.geometry,
            filter_map=FILTER_MAP,
        )
        self.assertIsInstance(f, GeometryIntersectionFilter)

    def test_intersection_geometry_filter_bbox_array(self):
        f = GeometryIntersectionFilter("cell_coords", BboxArrayField(), self.geometry)
        filtered = f.filter({"cell_coords": np.array([[0, 0, 1, 1], [2, 2, 3, 3]]).T})
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])

    def test_intersection_geometry_filter_bbox_array_line_bbox(self):
        # the BBOX of the line intersects all cells, but the line intersects
        # only 3 cells.
        line = LineString([(0.5, 0.5), (1.5, 0.5), (1.5, 1.5)])
        values = np.array([[0, 0, 1, 1], [0, 1, 0, 1], [1, 1, 2, 2], [1, 2, 1, 2]])
        f = GeometryIntersectionFilter("cell_coords", BboxArrayField(), line)
        filtered = f.filter({"cell_coords": values})
        self.assertTrue(filtered[0])
        self.assertTrue(filtered[2])
        self.assertTrue(filtered[3])
        self.assertFalse(filtered[1])

    def test_intersection_geometry_filter_bbox_array_line_bbox_nan(self):
        # the BBOX of the line intersects all cells, but the line intersects
        # only 3 cells.
        line = LineString([(0.5, 0.5), (1.5, 0.5), (1.5, 1.5)])
        values = np.array(
            [[np.nan, 0, 1, 1], [np.nan, 1, 0, 1], [np.nan, 1, 2, 2], [np.nan, 2, 1, 2]]
        )
        f = GeometryIntersectionFilter("cell_coords", BboxArrayField(), line)
        filtered = f.filter({"cell_coords": values})
        self.assertFalse(filtered[0])
        self.assertTrue(filtered[2])
        self.assertTrue(filtered[3])
        self.assertFalse(filtered[1])

    def test_intersection_geometry_filter_line_array(self):
        f = GeometryIntersectionFilter("line_coords", LineArrayField(), self.geometry)
        filtered = f.filter({"line_coords": np.array([[0, 0, 1, 1], [2, 2, 3, 3]]).T})
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])

    def test_intersection_geometry_filter_point_array(self):
        f = GeometryIntersectionFilter("coordinates", PointArrayField(), self.geometry)
        filtered = f.filter({"coordinates": np.array([[0.5, 2], [0.5, 2]])})
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])

    def test_intersection_geometry_filter_polygon_array(self):
        f = GeometryIntersectionFilter(
            "coordinates", PolygonArrayField(), self.geometry
        )
        filtered = f.filter(
            {
                "coordinates": np.array(
                    [  # + sign shows the split between x, y coords
                        np.array([0, 0, 2, 2] + [0, 2, 2, 0]),
                        np.array([5, 6, 6] + [5, 6, 5]),
                    ],
                    dtype=object,
                )
            }
        )
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])

    def test_intersection_geometry_filter_multiline_array(self):
        f = GeometryIntersectionFilter("coords", MultiLineArrayField(), self.geometry)
        filtered = f.filter(
            {
                "coords": np.array(
                    [  # + sign shows the split between x, y coords
                        np.array([0, 0.5, 2, 2] + [0, 0.5, 2, 0]),
                        np.array([5, 6, 6] + [5, 6, 5]),
                    ],
                    dtype=object,
                )
            }
        )
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])

    def test_intersection_geometry_filter_point_geom(self):
        geometry = Point(0.5, 0.5)
        f = GeometryIntersectionFilter("coordinates", PointArrayField(), geometry)
        filtered = f.filter({"coordinates": np.array([[0.5, 2], [0.5, 2]])})
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])

    def test_intersection_geometry_filter_line_geom(self):
        geometry = LineString([(0.5, 0.5), (1.5, 1.5)])
        f = GeometryIntersectionFilter("cell_coords", BboxArrayField(), geometry)
        filtered = f.filter({"cell_coords": np.array([[0, 0, 1, 1], [2, 2, 3, 3]]).T})
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

import numpy as np

from threedigrid.admin import constants
from threedigrid.admin.nodes.subsets import NODE_TYPE__IN_SUBSETS
from threedigrid.admin.prepare_utils import (
    add_or_update_datasets,
    db_objects_to_numpy_array_dict,
)


def as_numpy_array(array):
    if hasattr(array, "value"):
        return array.value
    return array[()]


class PrepareNodes:
    @staticmethod
    def get_node_pks(mapping1d, id_mapper):
        # pk template
        type_code = constants.TYPE_CODE_MAP["v2_connection_nodes"]

        content_pk = np.zeros(mapping1d["nend1d"].shape, dtype="i4")

        # with argsort
        sort_idx = np.argsort(mapping1d["nend1d"])
        idx = sort_idx[
            np.searchsorted(
                mapping1d["nend1d"],
                id_mapper.id_mapping[id_mapper.obj_slices[type_code]]["seq_id"],
                sorter=sort_idx,
            )
        ]

        content_pk[idx] = id_mapper.id_mapping[id_mapper.obj_slices[type_code]]["pk"]
        return content_pk

    @classmethod
    def prepare_datasource(cls, datasource, mapping1d, id_mapper, has_1d):
        if "id" not in list(datasource.keys()):
            datasource.set("id", np.arange(0, datasource["x_coordinate"].size))

        if "content_pk" not in list(datasource.keys()) and has_1d:
            datasource.set("content_pk", cls.get_node_pks(mapping1d, id_mapper))

        if "seq_id" not in list(datasource.keys()) and has_1d:
            datasource.set("seq_id", mapping1d["nend1d"][:])

        if "coordinates" not in list(datasource.keys()):
            datasource.set(
                "coordinates",
                np.array(
                    [
                        as_numpy_array(datasource["x_coordinate"]),
                        as_numpy_array(datasource["y_coordinate"]),
                    ]
                ),
            )


class PrepareConnectionNodes:
    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        node_group = h5py_file["nodes"]
        content_pk = node_group["content_pk"][:]

        connection_nodes_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.connection_nodes,
            ["pk", "initial_waterlevel", "storage_area"],
        )

        add_or_update_datasets(
            node_group,
            connection_nodes_numpy_array_dict,
            ["initial_waterlevel", "storage_area"],
            connection_nodes_numpy_array_dict["pk"],
            content_pk,
        )


class PrepareCells:
    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        node_group = h5py_file["nodes"]
        lgrmin = h5py_file["meta"]["lgrmin"][()]
        nodk = h5py_file["grid_coordinate_attributes"]["nodk"][()]
        nodm = h5py_file["grid_coordinate_attributes"]["nodm"][()]
        nodn = h5py_file["grid_coordinate_attributes"]["nodn"][()]
        ip = h5py_file["grid_coordinate_attributes"]["ip"][()]
        jp = h5py_file["grid_coordinate_attributes"]["jp"][()]

        node_types = h5py_file["nodes"]["node_type"][:]

        pixel_width = np.zeros(nodk.shape, dtype="int")
        pixel_coords = np.full((4, nodk.shape[0]), -9999, dtype="int")
        for node_type_subset in NODE_TYPE__IN_SUBSETS["2D_ALL"]:
            mask = node_types == node_type_subset
            pixel_width[mask] = lgrmin * 2 ** (nodk[mask] - 1)
            pixel_coords[0, mask] = ip[0, nodm[mask] - 1, nodk[mask] - 1] - 1
            pixel_coords[1, mask] = jp[0, nodn[mask] - 1, nodk[mask] - 1] - 1
            pixel_coords[2, mask] = ip[3, nodm[mask] - 1, nodk[mask] - 1]
            pixel_coords[3, mask] = jp[3, nodn[mask] - 1, nodk[mask] - 1]

        node_group.create_dataset("pixel_width", data=pixel_width, dtype="int")
        node_group.create_dataset("pixel_coords", data=pixel_coords, dtype="int")


class PrepareManholes:
    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        node_group = h5py_file["nodes"]
        content_pk = node_group["content_pk"][:]

        manhole_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.v2_manholes,
            [
                "pk",
                "surface_level",
                "display_name",
                "bottom_level",
                "calculation_type",
                "shape",
                "drain_level",
                "width",
                "manhole_indicator",
                "zoom_category",
                "connection_node_pk",
            ],
        )

        # extra field to distinguish manholes from connection nodes.
        is_manhole = np.full(len(threedi_datasource.v2_manholes), True)
        manhole_numpy_array_dict["is_manhole"] = is_manhole

        add_or_update_datasets(
            node_group,
            manhole_numpy_array_dict,
            [
                "surface_level",
                "display_name",
                "bottom_level",
                "calculation_type",
                "shape",
                "drain_level",
                "width",
                "manhole_indicator",
                "zoom_category",
                "is_manhole",
            ],
            manhole_numpy_array_dict["connection_node_pk"],
            content_pk,
        )



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

import json

import numpy as np


class NumpyEncoder(json.JSONEncoder):
    """
    convert specific numpy types to native python types
    """

    def default(self, obj):
        """
        overwrites the default method. Numpy integers and floats will be
        converted to native python types, numpy.ndarrays to python lists
        :param obj: object to encode
        :return: encoded object
        """
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, bytes):
            return obj.decode("utf-8")
        else:
            return super().default(obj)



import os
from typing import List

import pytest
from numpy.testing import assert_array_equal

from threedigrid.admin.gridresultadmin import GridH5StructureControl
from threedigrid.admin.structure_controls.exporters import (
    structure_control_actions_to_csv,
)
from threedigrid.admin.structure_controls.models import StructureControl

test_file_dir = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "test_files/structure_controls"
)
result_file = os.path.join(test_file_dir, "structure_control_actions_3di.nc")
grid_file = os.path.join(test_file_dir, "gridadmin.h5")


@pytest.fixture()
def gsc():
    gsc = GridH5StructureControl(grid_file, result_file)
    yield gsc
    gsc.close()


def test_group_by_id(gsc: GridH5StructureControl):
    struct_cntrl: StructureControl = gsc.table_control.group_by_id(
        "tablepumpcapacity_3_0_3153600000"
    )

    assert struct_cntrl.id == "tablepumpcapacity_3_0_3153600000"
    assert struct_cntrl.source_table == "v2_pumpstation"
    assert struct_cntrl.source_table_id == 2
    assert_array_equal(struct_cntrl.time, [0, 202, 602, 1202, 2002])
    assert struct_cntrl.action_type == "set_pump_capacity"
    assert_array_equal(struct_cntrl.action_value_1, [0.0, 0.1, 0.0, 0.2, 0.0])
    assert_array_equal(struct_cntrl.action_value_2, [0.0, 0.0, 0.0, 0.0, 0.0])
    assert_array_equal(struct_cntrl.is_active, [1, 1, 1, 1, 1])


def test_group_by_id_0(gsc: GridH5StructureControl):
    struct_cntrl: StructureControl = gsc.table_control.group_by_id("x")
    assert struct_cntrl is None


def test_group_by_action_type(gsc: GridH5StructureControl):
    struct_cntrls: List[StructureControl] = gsc.table_control.group_by_action_type(
        "set_pump_capacity"
    )
    assert len(struct_cntrls) == 2
    assert isinstance(struct_cntrls[0], StructureControl)
    assert struct_cntrls[0].id == "tablepumpcapacity_2_0_3153600000"
    assert struct_cntrls[1].id == "tablepumpcapacity_3_0_3153600000"


def test_group_by_is_active(gsc: GridH5StructureControl):
    struct_cntrls: List[StructureControl] = gsc.table_control.group_by_is_active(0)
    assert len(struct_cntrls) == 1
    assert struct_cntrls[0].id == "tablecrestlevel_1_0_3153600000"


def test_group_by_time(gsc: GridH5StructureControl):
    struct_cntrls: List[StructureControl] = gsc.table_control.group_by_time(241, 599)
    assert len(struct_cntrls) == 2
    assert struct_cntrls[0].id == "tablecrestlevel_1_0_3153600000"
    assert struct_cntrls[1].id == "tablepumpcapacity_2_0_3153600000"


def test_group_by_time_incorrect(gsc: GridH5StructureControl):
    struct_cntrls: List[StructureControl] = gsc.table_control.group_by_time(10, 0)
    assert len(struct_cntrls) == 0


def test_group_by_action_value(gsc: GridH5StructureControl):
    struct_cntrls: List[StructureControl] = gsc.table_control.group_by_action_value_1(
        0, 0.1
    )
    assert len(struct_cntrls) == 3
    assert struct_cntrls[0].id == "tablecrestlevel_1_0_3153600000"
    assert struct_cntrls[1].id == "tablepumpcapacity_2_0_3153600000"
    assert struct_cntrls[2].id == "tablepumpcapacity_3_0_3153600000"


def test_group_by_action_value_2(gsc: GridH5StructureControl):
    struct_cntrls: List[StructureControl] = gsc.table_control.group_by_action_value_2(
        0, 0
    )
    assert len(struct_cntrls) == 3
    assert struct_cntrls[0].id == "tablecrestlevel_1_0_3153600000"
    assert struct_cntrls[1].id == "tablepumpcapacity_2_0_3153600000"
    assert struct_cntrls[2].id == "tablepumpcapacity_3_0_3153600000"


def test_group_by_action_value_negative(gsc: GridH5StructureControl):
    struct_cntrls: List[StructureControl] = gsc.table_control.group_by_action_value_1(
        -1, -0.1
    )
    assert len(struct_cntrls) == 1
    assert struct_cntrls[0].id == "tablecrestlevel_1_0_3153600000"


def test_export_method(gsc):
    csv_path = os.path.join(test_file_dir, "test_struct_control_actions.csv")
    structure_control_actions_to_csv(gsc, csv_path)
    assert os.path.exists(csv_path)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

from threedigrid.admin.exporter_constants import DEFAULT_EXPORT_FIELDS
from threedigrid.admin.exporters.geopackage.exporter import GpkgExporter
from threedigrid.admin.serializers import GeoJsonSerializer
from threedigrid.orm.base.models import Model as BaseModel
from threedigrid.orm.fields import GeomArrayField, LineArrayField
from threedigrid.orm.filters import FILTER_MAP

from .constants import GEOJSON_DRIVER_NAME, SHP_DRIVER_NAME


class Model(BaseModel):
    _filter_map = FILTER_MAP

    def to_centroid(self):
        """
        Returns: the filtered values with geometries
                 transformed to centroids
        """
        selection = self.data
        for field_name in self._field_names:
            field = self._get_field(field_name)
            if isinstance(field, LineArrayField) and field_name in selection:
                # Use the to_centroid function of the GeomArrayField
                # to allow different geometry to be "centroided"
                selection[field_name] = field.to_centroid(selection[field_name])
        return selection

    def _is_coords(self, field_name):
        return isinstance(self._get_field(field_name), GeomArrayField)

    def _includes_coords(self, selection):
        """
        Returns: true if selection contains any data that
                 has coordinates.
        """
        found = False
        for field_name in self._field_names:
            if isinstance(self._get_field(field_name), GeomArrayField):
                found = True
                break
        return found

    def reproject_to(self, target_epsg_code):
        """
        Returns: a new class including the reprojection.

        Note: the last reproject_to is used:
            reproject_to('28992').reproject_to('4326') == reproject_to('4326')
        """
        new_class_kwargs = dict(self.class_kwargs)
        new_class_kwargs.update({"reproject_to_epsg": target_epsg_code})

        return self.__init_class(self.__class__, **new_class_kwargs)

    def _do_reproject_value(self, value, field_name, target_epsg_code):
        if target_epsg_code == self.epsg_code:
            # Already done
            return value

        field = self._get_field(field_name)

        return field.reproject(value, self.epsg_code, target_epsg_code)

    def __do_reproject(self, selection, target_epsg_code):
        """
        Reproject coordinate fields in selection to the
        target_epsg_code
        """
        if target_epsg_code == self.epsg_code:
            # Already done
            return selection

        # Reproject all GeomDataField's
        for field_name in self._field_names:
            field = self._get_field(field_name)
            if isinstance(field, GeomArrayField) and field_name in selection:
                # Use the reproject function of the GeomDataField
                # to allow different geometry to be reprojected
                selection[field_name] = field.reproject(
                    selection[field_name], self.epsg_code, target_epsg_code
                )
        return selection

    def to_shape(self, file_name, **kwargs):
        self._to_ogr(SHP_DRIVER_NAME, file_name, **kwargs)

    def to_gpkg(
        self,
        file_name,
        layer_name=None,
        field_definitions=None,
        progress_func=None,
        **kwargs
    ):
        # By default use class name in lowercase as layer_name
        if layer_name is None:
            layer_name = self.__class__.__name__.lower()
            if layer_name.endswith("s"):
                layer_name = layer_name[:-1]

        if field_definitions is None:
            if not hasattr(self, "GPKG_DEFAULT_FIELD_MAP"):
                raise Exception(
                    "field_definitions is not defined and no default field map present"
                )

            field_definitions = self.GPKG_DEFAULT_FIELD_MAP
        exporter = GpkgExporter(self)
        exporter.save(
            file_name,
            layer_name=layer_name,
            field_definitions=field_definitions,
            progress_func=progress_func,
            **kwargs
        )

    def to_geojson(self, file_name, **kwargs):
        if kwargs.get("use_ogr", False):
            self._to_ogr(GEOJSON_DRIVER_NAME, file_name, **kwargs)
        else:
            fields = kwargs.get(
                "fields", DEFAULT_EXPORT_FIELDS[self.__class__.__name__]
            )
            if fields == "ALL":
                fields = self._field_names

            indent = kwargs.get("indent", None)
            serializer = GeoJsonSerializer(fields, self, indent)
            serializer.save(file_name)

    def _to_ogr(self, driver_name, file_name, **kwargs):
        exporter = self._get_exporter(driver_name)
        if not exporter:
            raise AttributeError(
                "Instance {} has no {} exporter".format(self, driver_name)
            )
        exporter.set_driver(driver_name=driver_name)
        epsg_code = self.epsg_code
        if self.reproject_to_epsg:
            epsg_code = self.reproject_to_epsg

        exporter.save(file_name, self.data, epsg_code, **kwargs)

    def _get_exporter(self, driver_name):
        for exporter in self._exporters:
            if driver_name in exporter.supported_drivers:
                return exporter
        return None



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

import h5py
import numpy as np

DT_VARIABLE = h5py.special_dtype(vlen=np.dtype("float64"))


class PrepareLevees:
    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        v2_levee_instances = threedi_datasource.levees
        group_name = "levees"
        if group_name not in h5py_file:
            gr = h5py_file.create_group(group_name)
        else:
            gr = h5py_file["levees"]

        dset_coords = gr.create_dataset(
            "coords", (len(v2_levee_instances),), dtype=DT_VARIABLE
        )
        dset_clevel = gr.create_dataset(
            "crest_level", (len(v2_levee_instances),), dtype="f8"
        )
        dset_mbreach_depth = gr.create_dataset(
            "max_breach_depth", (len(v2_levee_instances),), dtype="f8"
        )
        dset_id = gr.create_dataset("id", (len(v2_levee_instances),), dtype="i4")
        for i, levee in enumerate(v2_levee_instances):
            # "F" means to flatten in column-major (Fortran- style) order
            dset_coords[i] = np.array(levee.the_geom.coords).flatten("F")
            dset_clevel[i] = levee.crest_level
            dset_mbreach_depth[i] = levee.max_breach_depth or -9999
            dset_id[i] = levee.pk



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

try:
    from osgeo import ogr
except ImportError:
    ogr = None

import numpy as np

from threedigrid.admin import constants
from threedigrid.geo_utils import raise_import_exception


def as_numpy_array(array):
    if hasattr(array, "value"):
        return array.value
    return array[()]


class PrepareBreaches:
    """
    prepares breaches for visualization
    """

    def __init__(self):
        if ogr is None:
            raise_import_exception("ogr")

    @staticmethod
    def get_coordinates(levees, line_coords, levl):
        breaches_x = np.zeros(levl.shape, dtype="f8")
        breaches_y = np.zeros(levl.shape, dtype="f8")

        for i, line_id in enumerate(levl):
            if i == 0:
                continue
            line = ogr.Geometry(ogr.wkbLineString)

            line.AddPoint(line_coords[0][line_id], line_coords[1][line_id])
            line.AddPoint(line_coords[2][line_id], line_coords[3][line_id])

            for levee_geom in levees.geoms:
                if not levee_geom.Intersect(line):
                    continue
                intersection = levee_geom.Intersection(line)
                breaches_x[i] = intersection.GetX()
                breaches_y[i] = intersection.GetY()
                break

        return np.array([breaches_x, breaches_y])

    @classmethod
    def prepare_datasource(cls, datasource, kcu, id_mapper, levees, line_coords):
        """
        :param datasource: datasource for breaches like HFD5 group
        :param kcu: array of kcu values for datasource
        :param id_mapper: threedigrid.admin.idmapper.IdMapper instance
        :param levees: threedigrid.admin.levees.model.Levees instance
        :param line_coords: coordinates from Lines instance
        :return:
        """

        # TODO: Check values below
        if "id" not in list(datasource.keys()):
            datasource.set("id", np.arange(0, datasource["levl"].size))

        if "seq_ids" not in list(datasource.keys()):
            datasource.set("seq_ids", np.arange(0, datasource["levl"].size))
        if "content_pk" not in list(datasource.keys()):
            content_pk = np.zeros(datasource["levl"].shape, dtype="i4")
            type_code = constants.TYPE_CODE_MAP["v2_breach"]
            id_mapping = id_mapper.id_mapping
            src = id_mapper.obj_slices[type_code]
            sort_by = as_numpy_array(datasource["seq_ids"])
            sort_idx = np.argsort(sort_by)
            idx = sort_idx[
                np.searchsorted(sort_by, id_mapping[src]["seq_id"], sorter=sort_idx)
            ]

            content_pk[idx] = id_mapping[src]["pk"]
            datasource.set("content_pk", content_pk)

        levl = as_numpy_array(datasource["levl"])

        if "kcu" not in list(datasource.keys()):
            datasource.set("kcu", as_numpy_array(kcu)[levl])

        if "coordinates" not in list(datasource.keys()) and levees:
            datasource.set(
                "coordinates",
                cls.get_coordinates(levees, as_numpy_array(line_coords), levl),
            )



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

from threedigrid.orm.base.options import ModelMeta
from threedigrid.orm.base.timeseries_mixin import AggregateResultMixin, ResultMixin

BASE_COMPOSITE_FIELDS = {
    "au": ["Mesh2D_au", "Mesh1D_au"],
    "u1": ["Mesh2D_u1", "Mesh1D_u1"],
    "q": ["Mesh2D_q", "Mesh1D_q"],
    "qp": ["Mesh2D_qp"],
    "up1": ["Mesh2D_up1"],
    "_mesh_id": ["Mesh2DLine_id", "Mesh1DLine_id"],  # private
}

BASE_SUBSET_FIELDS = {
    "qp": {"2d_all": "Mesh2D_qp"},
    "up1": {"2d_all": "Mesh2D_up1"},
    "breach_depth": {"1d_all": "Mesh1D_breach_depth"},
    "breach_width": {"1d_all": "Mesh1D_breach_width"},
}


class LinesResultsMixin(ResultMixin):
    class Meta:
        # attributes for the given fields
        field_attrs = ["units", "long_name", "standard_name"]

        composite_fields = BASE_COMPOSITE_FIELDS
        subset_fields = BASE_SUBSET_FIELDS

        lookup_fields = ("id", "_mesh_id")

    def __init__(self, **kwargs):
        """Instantiate a line with netcdf results.

        Variables stored in the netcdf and related to lines are dynamically
        added as attributes as TimeSeriesArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super().__init__(**kwargs)


class LinesAggregateResultsMixin(AggregateResultMixin):
    class Meta(metaclass=ModelMeta):
        field_attrs = ["units", "long_name"]

        base_composition = BASE_COMPOSITE_FIELDS
        base_subset_fields = BASE_SUBSET_FIELDS

        composition_vars = {
            "au": ["min", "max", "avg"],
            "u1": ["min", "max", "avg"],
            "q": ["min", "max", "avg", "cum", "cum_positive", "cum_negative"],
            "up1": ["min", "max", "avg"],
            "qp": ["min", "max", "avg", "cum", "cum_positive", "cum_negative"],
        }

        lookup_fields = ("id", "_mesh_id")

    def __init__(self, **kwargs):
        """Instantiate a line with netcdf results.

        Variables stored in the netcdf and related to lines are dynamically
        added as attributes as TimeSeriesArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super().__init__(**kwargs)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
"""
Exporters
---------
"""

import logging
import os

try:
    from osgeo import ogr
except ImportError:
    ogr = None

try:
    from shapely import geometry as shapely_geom
except ImportError:
    shapely_geom = None


from threedigrid.admin import exporter_constants as const
from threedigrid.admin.constants import LINE_BASE_FIELDS_ALL, LINE_FIELD_NAME_MAP
from threedigrid.admin.utils import KCUDescriptor
from threedigrid.geo_utils import get_spatial_reference, raise_import_exception
from threedigrid.orm.base.exporters import BaseOgrExporter

logger = logging.getLogger(__name__)


class LinesOgrExporter(BaseOgrExporter):
    """
    Exports to ogr formats. You need to set the driver explicitly
    before calling save()
    """

    def __init__(self, lines):
        """
        :param lines: lines.models.Lines instance
        """
        self._lines = lines
        self.supported_drivers = {
            const.GEO_PACKAGE_DRIVER_NAME,
            const.SHP_DRIVER_NAME,
            const.GEOJSON_DRIVER_NAME,
        }
        self.driver = None

    def save(self, file_name, line_data, target_epsg_code, **kwargs):
        """
        save to file format specified by the driver, e.g. shapefile

        :param file_name: name of the outputfile
        :param line_data: dict of line data
        """
        if self.driver is None:
            self.set_driver(extension=os.path.splitext(file_name)[1])

        assert self.driver is not None

        kcu_dict = KCUDescriptor()
        geomtype = 0
        sr = get_spatial_reference(target_epsg_code)

        geom_source = "from_threedicore"
        if "geom" in kwargs:
            geom_source = kwargs["geom"]

        # shapely is needed for the LineString creation, check if installed
        if geom_source == "from_spatialite" and shapely_geom is None:
            raise_import_exception("shapely")

        self.del_datasource(file_name)
        data_source = self.driver.CreateDataSource(file_name)
        layer = data_source.CreateLayer(str(os.path.basename(file_name)), sr, geomtype)
        fields = kwargs.get("fields", LINE_BASE_FIELDS_ALL)
        for field_name, field_type in fields.items():
            layer.CreateField(
                ogr.FieldDefn(str(field_name), const.OGR_FIELD_TYPE_MAP[field_type])
            )
        _definition = layer.GetLayerDefn()

        node_a = line_data["line"][0]
        node_b = line_data["line"][1]
        for i in range(node_a.size):
            if line_data["id"][i] == 0:
                continue  # skip the dummy element
            if geom_source == "from_threedicore":
                line = ogr.Geometry(ogr.wkbLineString)
                line.AddPoint(
                    line_data["line_coords"][0][i], line_data["line_coords"][1][i]
                )
                line.AddPoint(
                    line_data["line_coords"][2][i], line_data["line_coords"][3][i]
                )
            elif geom_source == "from_spatialite":
                linepoints = (
                    line_data["line_geometries"][i].reshape(2, -1).T.astype("float64")
                )
                line_geom = shapely_geom.LineString(linepoints)
                line = ogr.CreateGeometryFromWkt(line_geom.wkt)

            feature = ogr.Feature(_definition)
            feature.SetGeometry(line)
            for field_name, field_type in fields.items():
                fname = LINE_FIELD_NAME_MAP.get(field_name, field_name)
                if field_name == "kcu_descr":
                    raw_value = ""
                    try:
                        raw_value = str(kcu_dict[int(line_data["kcu"][i])])
                    except KeyError:
                        pass
                elif field_name == "node_a":
                    raw_value = node_a[i]
                elif field_name == "node_b":
                    raw_value = node_b[i]
                else:
                    try:
                        raw_value = line_data[fname][i]
                    except IndexError:
                        raw_value = None

                self.set_field(feature, field_name, field_type, raw_value)

            layer.CreateFeature(feature)
            feature.Destroy()



import os
import shutil
import tempfile
import unittest

import numpy as np
from osgeo import ogr
from pyproj import CRS

from threedigrid.admin.breaches.exporters import BreachesOgrExporter
from threedigrid.admin.constants import (
    NO_DATA_VALUE,
    SUBSET_1D_ALL,
    SUBSET_2D_OPEN_WATER,
)
from threedigrid.admin.gridadmin import GridH5Admin
from threedigrid.admin.lines.exporters import LinesOgrExporter
from threedigrid.admin.nodes.exporters import CellsOgrExporter, NodesOgrExporter

test_file_dir = os.path.join(os.getcwd(), "tests/test_files")

# the testfile is a copy of the v2_bergermeer gridadmin file
grid_admin_h5_file = os.path.join(test_file_dir, "gridadmin_v2.h5")


class GridAdminTest(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)

    def test_get_from_meta(self):
        self.assertIsNotNone(self.parser.get_from_meta("n2dtot"))
        self.assertIsNone(self.parser.get_from_meta("doesnotexist"))

    def test_get_extent_subset_onedee(self):
        extent_1D = self.parser.get_extent_subset(subset_name=SUBSET_1D_ALL)
        # should contain values
        self.assertTrue(np.any(extent_1D != NO_DATA_VALUE))

    def test_get_extent_subset_reproject(self):
        extent_1D = self.parser.get_extent_subset(subset_name=SUBSET_1D_ALL)
        extent_1D_proj = self.parser.get_extent_subset(
            subset_name=SUBSET_1D_ALL, target_epsg_code="4326"
        )
        self.assertTrue(np.all(extent_1D != extent_1D_proj))

    def test_get_extent_subset_twodee(self):
        extent_2D = self.parser.get_extent_subset(subset_name=SUBSET_2D_OPEN_WATER)
        # should contain values
        self.assertTrue(np.any(extent_2D != NO_DATA_VALUE))

    def test_get_extent_subset_combi(self):
        extent_1D = self.parser.get_extent_subset(subset_name=SUBSET_1D_ALL)
        extent_2D = self.parser.get_extent_subset(subset_name=SUBSET_2D_OPEN_WATER)
        # should be different
        self.assertTrue(np.any(np.not_equal(extent_1D, extent_2D)))

    def test_get_model_extent(self):
        model_extent = self.parser.get_model_extent()
        np.testing.assert_almost_equal(
            model_extent,
            np.array([105427.6, 511727.0515702, 115887.0, 523463.3268483]),
            decimal=3,
        )

    def test_get_model_extent_extra_extent(self):
        onedee_extra = np.array([100000.0, 90000.0, 550000.0, 580000.0])

        extra_extent = {"extra_extent": [onedee_extra]}
        model_extent = self.parser.get_model_extent(**extra_extent)
        np.testing.assert_equal(
            model_extent, np.array([100000.0, 90000.0, 550000.0, 580000.0])
        )

    def test_get_model_extent_extra_extent2(self):
        onedee_extra = np.array([106666.6, 106666.6, 550000.0, 580000.0])

        extra_extent = {"extra_extent": [onedee_extra]}
        model_extent = self.parser.get_model_extent(**extra_extent)
        np.testing.assert_almost_equal(
            model_extent, np.array([105427.6, 106666.6, 550000.0, 580000.0]), decimal=3
        )

    def test_properties(self):
        self.assertTrue(hasattr(self.parser, "has_groundwater"))
        self.assertTrue(hasattr(self.parser, "has_levees"))
        self.assertTrue(hasattr(self.parser, "threedicore_version"))
        self.assertTrue(hasattr(self.parser, "has_1d"))
        self.assertTrue(hasattr(self.parser, "has_2d"))
        self.assertTrue(hasattr(self.parser, "epsg_code"))
        self.assertTrue(hasattr(self.parser, "revision_hash"))
        self.assertTrue(hasattr(self.parser, "revision_nr"))
        self.assertTrue(hasattr(self.parser, "model_slug"))

    def test_crs(self):
        crs = self.parser.crs
        assert isinstance(crs, CRS)
        assert crs.to_epsg() == 28992 or "28992" in crs.srs


class GridAdminLinesTest(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)
        d = tempfile.mkdtemp()
        self.f = os.path.join(d, "test_lines.shp")

    def tearDown(self):
        shutil.rmtree(os.path.dirname(self.f))

    def test_fields(self):
        # Check dtype names
        assert set(self.parser.lines._meta.get_fields().keys()) == {
            "content_pk",
            "content_type",
            "id",
            "kcu",
            "lik",
            "line",
            "dpumax",
            "flod",
            "flou",
            "cross1",
            "cross2",
            "ds1d",
            "line_coords",
            "cross_pix_coords",
            "cross_weight",
            "discharge_coefficient_negative",
            "discharge_coefficient_positive",
            "line_geometries",
            "sewerage",
            "sewerage_type",
            "invert_level_start_point",
            "invert_level_end_point",
            "zoom_category",
            "ds1d_half",
        }

    def test_exporters(self):
        self.assertEqual(len(self.parser.lines._exporters), 1)
        self.assertIsInstance(self.parser.lines._exporters[0], LinesOgrExporter)

    def test_export_to_shape(self):
        self.parser.lines.to_shape(self.f)
        self.assertTrue(os.path.exists, self.f)
        s = ogr.Open(self.f)
        lyr = s.GetLayer()
        self.assertEqual(lyr.GetFeatureCount(), self.parser.lines.id.size - 1)


class GridAdminGridTest(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)
        self.grid = self.parser.grid

    def test_fields(self):
        assert set(self.grid._meta.get_fields().keys()) == {
            "id",
            "nodk",
            "nodm",
            "nodn",
            "ip",
            "jp",
        }

    def test_dx(self):
        expected = np.array([20.0, 40.0, 80.0, 160.0])
        np.testing.assert_almost_equal(self.grid.dx, expected)
        np.testing.assert_almost_equal(self.grid.filter(id=1).dx, expected)

    def test_n2dtot(self):
        expected = 5374
        self.assertEqual(self.grid.n2dtot, expected)
        self.assertEqual(self.grid.filter(id=1).n2dtot, expected)

    def test_transform(self):
        self.assertEqual(self.grid.transform, (0.5, 0.0, 106314.0, 0.0, 0.5, 514912.0))

    @unittest.skip("TODO")
    def test_get_pixel_map(self):
        self.parser.grid.get_pixel_map()


class GridAdminNodeTest(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)
        d = tempfile.mkdtemp()
        self.f = os.path.join(d, "test_nodes.shp")

    def tearDown(self):
        shutil.rmtree(os.path.dirname(self.f))

    def test_fields(self):
        # Check fields
        assert set(self.parser.nodes._meta.get_fields(only_names=True)) == {
            "cell_coords",
            "content_pk",
            "coordinates",
            "id",
            "node_type",
            "calculation_type",
            "seq_id",
            "zoom_category",
            "is_manhole",
            "sumax",
            "drain_level",
            "storage_area",
            "dmax",
            "initial_waterlevel",
            "dimp",
        }

    def test_locations_2d(self):
        self.assertGreater(len(self.parser.nodes.locations_2d), 0)
        frst = self.parser.nodes.locations_2d[0]
        # should contain three elements
        self.assertEqual(len(frst), 3)

    def test_exporters(self):
        self.assertEqual(len(self.parser.nodes._exporters), 1)
        self.assertIsInstance(self.parser.nodes._exporters[0], NodesOgrExporter)

    def test_export_to_shape(self):
        self.parser.nodes.to_shape(self.f)
        self.assertTrue(os.path.exists, self.f)
        s = ogr.Open(self.f)
        lyr = s.GetLayer()
        self.assertEqual(lyr.GetFeatureCount(), self.parser.nodes.id.size - 1)


class GridAdminBreachTest(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)
        d = tempfile.mkdtemp()
        self.f = os.path.join(d, "test_breaches.shp")

    def tearDown(self):
        shutil.rmtree(os.path.dirname(self.f))

    def test_fields(self):
        # Check fields
        assert set(self.parser.breaches._meta.get_fields().keys()) == {
            "content_pk",
            "coordinates",
            "id",
            "kcu",
            "levbr",
            "levl",
            "levmat",
            "line_geometries",
            "seq_ids",
            "code",
            "display_name",
        }

    def test_exporters(self):
        self.assertEqual(len(self.parser.breaches._exporters), 1)
        self.assertIsInstance(self.parser.breaches._exporters[0], BreachesOgrExporter)

    def test_export_to_shape(self):
        self.parser.breaches.to_shape(self.f)
        self.assertTrue(os.path.exists, self.f)
        s = ogr.Open(self.f)
        lyr = s.GetLayer()
        self.assertEqual(lyr.GetFeatureCount(), self.parser.breaches.id.size - 1)


class GridAdminCellsTest(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)
        d = tempfile.mkdtemp()
        self.f = os.path.join(d, "test_cells.shp")

    def tearDown(self):
        shutil.rmtree(os.path.dirname(self.f))

    def test_fields(self):
        # should have also z_coordinate
        assert set(self.parser.cells._meta.get_fields().keys()) == {
            "cell_coords",
            "content_pk",
            "coordinates",
            "id",
            "calculation_type",
            "node_type",
            "seq_id",
            "z_coordinate",
            "zoom_category",
            "is_manhole",
            "sumax",
            "pixel_width",
            "pixel_coords",
            "drain_level",
            "storage_area",
            "dmax",
            "initial_waterlevel",
            "has_dem_averaged",
            "dimp",
        }

    def test_get_id_from_xy(self):
        # should yield tow ids, one for 2d, one for groundwater
        self.assertListEqual(self.parser.cells.get_id_from_xy(1.0, 2.0), [])
        # first coordinate pair + some offset
        x = self.parser.cells.coordinates[0][1] + 0.5
        y = self.parser.cells.coordinates[1][1] + 0.5
        self.assertListEqual(self.parser.cells.get_id_from_xy(x, y), [1, 5375])

    def test_get_id_from_xy_2d_open_water(self):
        self.assertListEqual(
            self.parser.cells.get_id_from_xy(1.0, 2.0, subset_name="2d_open_water"),
            [],
        )
        # first coordinate pair + some offset
        x = self.parser.cells.coordinates[0][1] + 0.5
        y = self.parser.cells.coordinates[1][1] + 0.5
        self.assertEqual(
            self.parser.cells.get_id_from_xy(x, y, subset_name="2d_open_water"),
            [1],
        )

    def test_get_id_from_xy_groundwater(self):
        self.assertListEqual(
            self.parser.cells.get_id_from_xy(1.0, 2.0, subset_name="groundwater_all"),
            [],
        )
        # first coordinate pair + some offset
        x = self.parser.cells.coordinates[0][1] + 0.5
        y = self.parser.cells.coordinates[1][1] + 0.5
        self.assertEqual(
            self.parser.cells.get_id_from_xy(x, y, subset_name="groundwater_all"),
            [5375],
        )

    def test_get_extent_pixels(self):
        cells = self.parser.cells
        assert cells.get_extent_pixels() == (0, 0, 9600, 9920)

    def test_iter_by_tile(self):
        cells = self.parser.cells.subset("2D_OPEN_WATER")
        w, h = 320 * 50, 1280
        tiles = list(cells.iter_by_tile(w, h))
        assert len(tiles) == 8  # ceil(9920/1280)
        total = 0
        for i, (bbox, cells) in enumerate(tiles):
            assert bbox == (0, i * h, w, (i + 1) * h)
            assert np.all(cells.pixel_coords[1] >= i * h)
            assert np.all(cells.pixel_coords[1] < (i + 1) * h)
            total += cells.count
        assert total == self.parser.cells.subset("2D_OPEN_WATER").count

    def test_iter_by_tile_subset_y(self):
        cells = self.parser.cells.subset("2D_OPEN_WATER").filter(
            pixel_coords__in_bbox=(0, 2000, 9600, 3000)
        )
        w, h = 320 * 50, 1280
        tiles = list(cells.iter_by_tile(w, h))
        assert len(tiles) == 2  # 1280 - 2560 and 2560 - 3840
        assert tiles[0][0] == (0, 1280, w, 1280 * 2)
        assert tiles[1][0] == (0, 1280 * 2, w, 1280 * 3)
        assert cells.count == (tiles[0][1].count + tiles[1][1].count)

    def test_iter_by_tile_subset_x(self):
        cells = self.parser.cells.subset("2D_OPEN_WATER").filter(
            pixel_coords__in_bbox=(2000, 0, 3000, 9920)
        )
        w, h = 1280, 320 * 50
        tiles = list(cells.iter_by_tile(w, h))
        assert len(tiles) == 2  # 1280 - 2560 and 2560 - 3840
        assert tiles[0][0] == (1280, 0, 1280 * 2, h)
        assert tiles[1][0] == (1280 * 2, 0, 1280 * 3, h)
        assert cells.count == (tiles[0][1].count + tiles[1][1].count)

    def test_iter_by_tile_should_match(self):
        cells = self.parser.cells.subset("2D_OPEN_WATER")
        for tile in ((321, 1280), (160, 1280), (500, 1280), (1280, 500)):
            with self.assertRaises(ValueError):
                list(cells.iter_by_tile(*tile))

    def test_exporters(self):
        self.assertEqual(len(self.parser.cells._exporters), 1)
        self.assertIsInstance(self.parser.cells._exporters[0], CellsOgrExporter)

    def test_export_to_shape(self):
        self.parser.cells.to_shape(self.f)
        self.assertTrue(os.path.exists, self.f)


class GridAdminCrossSectionsTest(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)
        d = tempfile.mkdtemp()
        self.f = os.path.join(d, "test_cross_sections.shp")

    def tearDown(self):
        shutil.rmtree(os.path.dirname(self.f))

    def test_fields(self):
        # Check dtype names
        assert set(self.parser.cross_sections._meta.get_fields().keys()) == {
            "id",
            "code",
            "shape",
            "content_pk",
            "width_1d",
            "offset",
            "count",
            "tables",
        }


class NodeFilterTests(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)

    def test_nodes_filter_id_eq(self):
        filtered = self.parser.nodes.filter(id=3).data["coordinates"]
        (trues,) = np.where(self.parser.nodes.data["id"] == 3)
        expected = self.parser.nodes.data["coordinates"][:, trues]
        np.testing.assert_equal(filtered, expected)

    def test_nodes_filter_id_ne(self):
        filtered = self.parser.nodes.filter(id__ne=3).data["coordinates"]
        (trues,) = np.where(self.parser.nodes.data["id"] != 3)
        expected = self.parser.nodes.data["coordinates"][:, trues]
        np.testing.assert_equal(filtered, expected)

    def test_nodes_filter_id_gt(self):
        filtered = self.parser.nodes.filter(id__gt=3).data["coordinates"]
        (trues,) = np.where(self.parser.nodes.data["id"] > 3)
        expected = self.parser.nodes.data["coordinates"][:, trues]
        np.testing.assert_equal(filtered, expected)

    def test_nodes_filter_id_gte(self):
        filtered = self.parser.nodes.filter(id__gte=3).data["coordinates"]
        (trues,) = np.where(self.parser.nodes.data["id"] >= 3)
        expected = self.parser.nodes.data["coordinates"][:, trues]
        np.testing.assert_equal(filtered, expected)

    def test_nodes_filter_id_lt(self):
        filtered = self.parser.nodes.filter(id__lt=3).data["coordinates"]
        (trues,) = np.where(self.parser.nodes.data["id"] < 3)
        expected = self.parser.nodes.data["coordinates"][:, trues]
        np.testing.assert_equal(filtered, expected)

    def test_nodes_filter_id_lte(self):
        filtered = self.parser.nodes.filter(id__lte=3).data["coordinates"]
        (trues,) = np.where(self.parser.nodes.data["id"] <= 3)
        expected = self.parser.nodes.data["coordinates"][:, trues]
        np.testing.assert_equal(filtered, expected)

    def test_nodes_filter_id_in(self):
        """Verify that 'in' filter returns the correct data."""
        filtered = self.parser.nodes.filter(id__in=list(range(3, 7))).data[
            "coordinates"
        ]
        (trues,) = np.where(
            (self.parser.nodes.data["id"] >= 3) & (self.parser.nodes.data["id"] < 7)
        )
        expected = self.parser.nodes.data["coordinates"][:, trues]
        np.testing.assert_equal(filtered, expected)

    # TODO: fix this
    @unittest.skip(
        "This test should succeed, but in our code the reprojected tile "
        "bounds go out of bounds, which causes errors. TODO: fix this."
    )
    def test_nodes_filter_id_in_tile(self):
        # at z=0 we have a single base tile
        filtered = self.parser.nodes.filter(coordinates__in_tile=[0, 0, 0]).data["id"]
        expected = self.parser.nodes.data["id"]
        self.assertTrue(len(filtered) != 0)
        np.testing.assert_equal(filtered, expected)

    # some special cases

    def test_nodes_filter_id_eq_chained(self):
        """Eq filter can be chained."""
        filtered = self.parser.nodes.filter(id=3).filter(id=3).data["coordinates"]
        (trues,) = np.where(self.parser.nodes.data["id"] == 3)
        expected = self.parser.nodes.data["coordinates"][:, trues]
        np.testing.assert_equal(filtered, expected)

    def test_nodes_chained_same_filter_id_in(self):
        """In filters can be chained."""
        filtered = (
            self.parser.nodes.filter(id__in=list(range(1, 10)))
            .filter(id__in=list(range(1, 10)))
            .filter(id__in=list(range(1, 10)))
            .data["coordinates"]
        )
        expected = self.parser.nodes.filter(id__in=list(range(1, 10))).data[
            "coordinates"
        ]
        np.testing.assert_equal(filtered, expected)

    def test_nodes_chained_filter_id_in(self):
        filtered = (
            self.parser.nodes.filter(id__in=list(range(1, 8)))
            .filter(id__in=list(range(3, 7)))
            .data["coordinates"]
        )
        expected = self.parser.nodes.filter(id__in=list(range(3, 7))).data[
            "coordinates"
        ]
        np.testing.assert_equal(filtered, expected)

    def test_nodes_chained_filter_id_in_2(self):
        filtered = (
            self.parser.nodes.filter(id__in=list(range(1, 10)))
            .filter(id__in=list(range(5, 20)))
            .data["coordinates"]
        )
        expected = self.parser.nodes.filter(id__in=list(range(5, 10))).data[
            "coordinates"
        ]
        np.testing.assert_equal(filtered, expected)

    def test_manhole_filter(self):
        non_manholes1 = self.parser.nodes.filter(is_manhole=False).count
        non_manholes2 = self.parser.nodes.filter(is_manhole=0).count
        non_manholes3 = self.parser.nodes.filter(is_manhole__eq=0).count
        non_manholes4 = self.parser.nodes.filter(is_manhole__eq=False).count
        non_manholes5 = self.parser.nodes.filter(is_manhole__ne=1).count
        non_manholes6 = self.parser.nodes.filter(is_manhole__ne=True).count
        self.assertEqual(non_manholes1, non_manholes2)
        self.assertEqual(non_manholes2, non_manholes3)
        self.assertEqual(non_manholes3, non_manholes4)
        self.assertEqual(non_manholes4, non_manholes5)
        self.assertEqual(non_manholes5, non_manholes6)

        manholes1 = self.parser.nodes.filter(is_manhole=True).count
        manholes2 = self.parser.nodes.filter(is_manhole=1).count
        manholes3 = self.parser.nodes.filter(is_manhole__eq=1).count
        manholes4 = self.parser.nodes.filter(is_manhole__eq=True).count
        manholes5 = self.parser.nodes.filter(is_manhole__ne=0).count
        manholes6 = self.parser.nodes.filter(is_manhole__ne=False).count
        self.assertEqual(manholes1, manholes2)
        self.assertEqual(manholes2, manholes3)
        self.assertEqual(manholes3, manholes4)
        self.assertEqual(manholes4, manholes5)
        self.assertEqual(manholes5, manholes6)

        self.assertEqual(manholes1 + non_manholes2, self.parser.nodes.count)

    def test_node_filter_keeps_has_1d(self):
        """Property has_1d doesn't disappear"""
        self.assertEqual(
            self.parser.nodes.has_1d, self.parser.nodes.filter(id=1).has_1d
        )


class LineFilterTests(unittest.TestCase):
    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)

    def test_lines_filter_id_eq(self):
        # from nose.tools import set_trace; set_trace()
        filtered = self.parser.lines.filter(id=3).data["line_coords"]
        (trues,) = np.where(self.parser.lines.data["id"] == 3)
        expected = self.parser.lines.data["line_coords"][:, trues]
        np.testing.assert_equal(filtered, expected)

    def test_lines_filter_id_in(self):
        filtered = self.parser.lines.filter(id__in=list(range(3, 7))).data[
            "line_coords"
        ]
        (trues,) = np.where(
            (self.parser.lines.data["id"] >= 3) & (self.parser.lines.data["id"] < 7)
        )
        expected = self.parser.lines.data["line_coords"][:, trues]
        np.testing.assert_equal(filtered, expected)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

import h5py
import numpy as np
from shapely import wkt
from shapely.geometry import MultiPoint, Point

from threedigrid.admin import constants
from threedigrid.admin.prepare_utils import (
    add_or_update_datasets,
    db_objects_to_numpy_array_dict,
)

DT_VARIABLE = h5py.special_dtype(vlen=np.dtype("float64"))


def as_numpy_array(array):
    if hasattr(array, "value"):
        return array.value
    return array[:]


class PrepareLines:
    @staticmethod
    def get_1d_object_info(datasource, id_mapper):
        """
        Set content type and content pk's based on id mapping and
        kcu's from calculationcore.

        :return: content_type and content_pk with length lines
        """

        # TODO channel might be different
        LINE_TYPES = [
            constants.TYPE_V2_PIPE,
            constants.TYPE_V2_CHANNEL,
            constants.TYPE_V2_CULVERT,
            constants.TYPE_V2_ORIFICE,
            constants.TYPE_V2_WEIR,
        ]

        _tmp_kcu = datasource["kcu"][:]
        filter_1d = (_tmp_kcu >= 0) & (_tmp_kcu <= 5)

        lik_all = as_numpy_array(datasource["lik"])
        lik_1d = as_numpy_array(datasource["lik"])[filter_1d]

        content_pk = np.zeros(lik_all.shape, dtype="i4")
        content_type = np.zeros(lik_all.shape, dtype="U32")

        _content_pk = np.zeros(lik_1d.shape, dtype="i4")
        _content_type = np.zeros(lik_1d.shape, dtype="U32")

        for line_type in LINE_TYPES:
            line_code = constants.TYPE_CODE_MAP[line_type]
            mapper_idx = id_mapper.obj_slices.get(line_code)
            if mapper_idx is None:
                continue
            seq_ids = id_mapper.id_mapping[mapper_idx]["seq_id"]
            for i in range(len(seq_ids)):
                slice1 = np.where(lik_1d == seq_ids[i])
                _content_pk[slice1] = id_mapper.id_mapping[mapper_idx]["pk"][i]
                _content_type[slice1] = line_type

        content_pk[filter_1d] = _content_pk
        content_type[filter_1d] = _content_type

        return content_pk, content_type

    @classmethod
    def make_line_geometries(cls, datasource, threedi_datasource):
        """

        create line geometries based on spatialite geometries but with
        segmentized for calculation core line segments
        :return:
        flattened array of line_geometries with length lines
        array[
            [x1, x2, x3, y1, y2, y3],
            [x1, x2, y1, y2]
        ]
        """

        DB_OBJECTS = [threedi_datasource.v2_channels, threedi_datasource.v2_culverts]

        LINE_TYPES = [constants.TYPE_V2_CHANNEL, constants.TYPE_V2_CULVERT]

        line_db_dict = dict(list(zip(LINE_TYPES, DB_OBJECTS)))

        size_array = as_numpy_array(datasource["lik"]).shape[0]
        line_geometries = np.full(size_array, 0, dtype=DT_VARIABLE)
        start_x = datasource["line_coords"][0][:]
        start_y = datasource["line_coords"][1][:]
        end_x = datasource["line_coords"][2][:]
        end_y = datasource["line_coords"][3][:]
        kcu = datasource["kcu"][:]
        xys = np.array(list(zip(start_x.T, end_x.T, start_y.T, end_y.T)))
        for i in range(len(line_geometries)):
            line_geometries[i] = np.array(xys[i])

        for line_type, db_objects in line_db_dict.items():
            for db_object in db_objects:
                line_idx = np.where(
                    (datasource["content_pk"][:] == db_object.pk)
                    & (datasource["content_type"][:] == line_type)
                )[0]
                geom = wkt.loads(db_object.the_geom.wkt)
                line_geometries[line_idx] = PrepareLines._cut_geometries(
                    geom,
                    start_x[line_idx],
                    start_y[line_idx],
                    end_x[line_idx],
                    end_y[line_idx],
                    kcu[line_idx],
                )

        return line_geometries

    @staticmethod
    def _cut_geometries(geom, start_x, start_y, end_x, end_y, kcu_array):
        """
        Segmentize line geometry based on start and end points from calc
        line segments

        :param geom: orginal geometry from the sqlite database
        :param start_x: threedicore x-coordinates for start points
        :param start_y: threedicore y-coordinates for start points
        :param end_x: threedicore x-coordinates for end points
        :param end_y: threedicore y-coordinates for end points
        :kcu_array: corresponding kcu values for threedicore coordinates

        :returns  an array of seperate line geometries for number of
        calc lines on geometry, like so::

            array[
                [x1, x2, x3, y1, y2, y3],
                [x1, x2, y1, y2]
            ]
        """

        cut_geometries = np.zeros((len(start_x),), dtype=DT_VARIABLE)
        start_points = MultiPoint(list(zip(start_x, start_y)))
        end_points = MultiPoint(list(zip(end_x, end_y)))

        for piece in range(len(cut_geometries)):
            kcu = kcu_array[piece]
            if kcu == 0:
                start_pnt = geom.interpolate(geom.project(start_points[piece]))
                end_pnt = geom.interpolate(geom.project(end_points[piece]))
            else:
                start_pnt = start_points[piece]
                end_pnt = end_points[piece]
            start_distance = round(geom.project(start_pnt), 3)
            end_distance = round(geom.project(end_pnt), 3)

            # Don't use the z-coordinate
            if geom.has_z:
                coords = [(x, y) for x, y, z in geom.coords]
            else:
                coords = list(geom.coords)
            start_set = False
            # no additional calc points
            if start_distance <= 0.0 and end_distance >= geom.length:
                linestring = np.array(
                    [(start_pnt.x, start_pnt.y)]
                    + coords[1:-1]
                    + [(end_pnt.x, end_pnt.y)]
                )
                # "F" means to flatten in column-major (Fortran- style) order
                cut_geometries[piece] = linestring.flatten("F")

            for i, p in enumerate(coords):
                # dist to vertex
                pd = round(geom.project(Point(p)), 3)

                # should not happen but sometimes drawing direction does not
                # correspond with start and endpoints, so flip them
                if start_distance > end_distance:
                    # This is not safe!!!!!!
                    # TODO: check if this actually can happen
                    start_distance, end_distance = end_distance, start_distance
                    start_points, end_points = end_points, start_points

                # check for start point
                if (pd == start_distance) and not start_set:
                    start_pnt = []
                    start_i = i
                    start_set = True
                # should not happen but in case pd (point) and first vertex
                # do not have the same position move pd back
                elif (pd > start_distance) and not start_set:
                    start_pnt = [(start_pnt.x, start_pnt.y)]
                    start_i = i
                    start_set = True
                if pd >= end_distance:
                    linestring = np.array(
                        start_pnt + coords[start_i:i] + [(end_pnt.x, end_pnt.y)]
                    )
                    # "F" means to flatten in column-major (Fortran- style)
                    # order
                    cut_geometries[piece] = linestring.flatten("F")
                    break

        return cut_geometries

    @classmethod
    def prepare_datasource(
        cls, datasource, id_mapper, threedi_datasource, node_coordinates, has_1d
    ):
        if "id" not in list(datasource.keys()):
            datasource.set("id", np.arange(0, datasource["kcu"].size))

        if has_1d and (
            "content_pk" not in list(datasource.keys())
            or "content_type" not in list(datasource.keys())
        ):
            content_pk, content_type = cls.get_1d_object_info(datasource, id_mapper)

            if "content_pk" not in list(datasource.keys()):
                datasource.set("content_pk", content_pk)
            if "content_type" not in list(datasource.keys()):
                datasource.set(
                    "content_type", [x.encode("ascii") for x in content_type]
                )

        if "line_coords" not in list(datasource.keys()):
            line = as_numpy_array(datasource["line"])
            x, y = as_numpy_array(node_coordinates[0]), as_numpy_array(
                node_coordinates[1]
            )
            datasource.set(
                "line_coords",
                np.array([x[line[0]], y[line[0]], x[line[1]], y[line[1]]]),
            )

        if has_1d and "line_geometries" not in list(datasource.keys()):
            line_geometries = cls.make_line_geometries(datasource, threedi_datasource)
            datasource.set("line_geometries", np.array(line_geometries))


class PrepareChannels:
    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        line_group = h5py_file["lines"]

        content_pk = line_group["content_pk"][:]
        content_type = line_group["content_type"][:]

        channels_field_names = [
            "pk",
            "code",
            "calculation_type",
            "dist_calc_points",
            "connection_node_start_pk",
            "connection_node_end_pk",
            "zoom_category",
        ]

        channels_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.v2_channels, channels_field_names
        )

        # discharge_coefficient defaults to 1.0 for channels
        channels_field_names.append("discharge_coefficient")
        channels_numpy_array_dict["discharge_coefficient"] = np.ones(
            len(content_pk), dtype=np.float32
        )

        add_or_update_datasets(
            line_group,
            channels_numpy_array_dict,
            channels_field_names,
            channels_numpy_array_dict["pk"],
            content_pk,
            ignore_mask=content_type != "v2_channel",
        )


class PreparePipes:
    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        line_group = h5py_file["lines"]
        content_pk = line_group["content_pk"][:]
        content_type = line_group["content_type"][:]

        pipes_field_names = [
            "pk",
            "display_name",
            "invert_level_start_point_raw",
            "invert_level_end_point_raw",
            "friction_type",
            "friction_value",
            "material",
            "sewerage_type",
            "calculation_type",
            "connection_node_start_pk",
            "connection_node_end_pk",
            "zoom_category",
            "cross_section_definition__db_width",
            "cross_section_definition__db_height",
            "cross_section_definition__db_shape",
        ]

        pipes_field_name_override = {
            "invert_level_start_point_raw": "invert_level_start_point",
            "invert_level_end_point_raw": "invert_level_end_point",
            "cross_section_definition__db_width": "cross_section_width",
            "cross_section_definition__db_height": "cross_section_height",
            "cross_section_definition__db_shape": "cross_section_shape",
        }

        pipes_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.v2_pipes, pipes_field_names
        )

        # discharge_coefficient defaults to 1.0 for pipes
        pipes_field_names.append("discharge_coefficient")
        pipes_numpy_array_dict["discharge_coefficient"] = np.ones(
            len(content_pk), dtype=np.float32
        )

        add_or_update_datasets(
            line_group,
            pipes_numpy_array_dict,
            pipes_field_names,
            pipes_numpy_array_dict["pk"],
            content_pk,
            ignore_mask=content_type != "v2_pipe",
            field_name_override=pipes_field_name_override,
        )


class PrepareWeirs:
    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        line_group = h5py_file["lines"]
        content_pk = line_group["content_pk"][:]
        content_type = line_group["content_type"][:]

        weirs_field_names = [
            "pk",
            "code",
            "display_name",
            "discharge_coefficient_negative",
            "discharge_coefficient_positive",
            "sewerage",
            "friction_type",
            "friction_value",
            "crest_type",
            "crest_level_raw",
            "connection_node_start_pk",
            "connection_node_end_pk",
            "zoom_category",
            "cross_section_definition__db_width",
            "cross_section_definition__db_height",
            "cross_section_definition__db_shape",
        ]

        weir_field_name_override = {
            "crest_level_raw": "crest_level",
            "cross_section_definition__db_width": "cross_section_width",
            "cross_section_definition__db_height": "cross_section_height",
            "cross_section_definition__db_shape": "cross_section_shape",
        }

        weirs_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.v2_weirs, weirs_field_names
        )

        add_or_update_datasets(
            line_group,
            weirs_numpy_array_dict,
            weirs_field_names,
            weirs_numpy_array_dict["pk"],
            content_pk,
            ignore_mask=content_type != "v2_weir",
            field_name_override=weir_field_name_override,
        )


class PrepareOrifices:
    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        line_group = h5py_file["lines"]
        content_pk = line_group["content_pk"][:]
        content_type = line_group["content_type"][:]

        orifices_field_names = [
            "pk",
            "display_name",
            "sewerage",
            "friction_type",
            "friction_value",
            "discharge_coefficient_negative",
            "discharge_coefficient_positive",
            "crest_type",
            "crest_level_raw",
            "connection_node_start_pk",
            "connection_node_end_pk",
            "zoom_category",
        ]

        orifices_field_name_override = {"crest_level_raw": "crest_level"}

        orifices_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.v2_orifices, orifices_field_names
        )

        add_or_update_datasets(
            line_group,
            orifices_numpy_array_dict,
            orifices_field_names,
            orifices_numpy_array_dict["pk"],
            content_pk,
            ignore_mask=content_type != "v2_orifice",
            field_name_override=orifices_field_name_override,
        )


class PrepareCulverts:
    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        line_group = h5py_file["lines"]
        content_pk = line_group["content_pk"][:]
        content_type = line_group["content_type"][:]

        culverts_field_names = [
            "pk",
            "code",
            "display_name",
            "discharge_coefficient_negative",
            "discharge_coefficient_positive",
            "friction_type",
            "friction_value",
            "invert_level_start_point_raw",
            "invert_level_end_point_raw",
            "calculation_type",
            "dist_calc_points",
            "connection_node_start_pk",
            "connection_node_end_pk",
            "zoom_category",
            "cross_section_definition__db_width",
            "cross_section_definition__db_height",
            "cross_section_definition__db_shape",
        ]

        culverts_field_name_override = {
            "invert_level_start_point_raw": "invert_level_start_point",
            "invert_level_end_point_raw": "invert_level_end_point",
            "cross_section_definition__db_width": "cross_section_width",
            "cross_section_definition__db_height": "cross_section_height",
            "cross_section_definition__db_shape": "cross_section_shape",
        }

        culverts_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.v2_culverts, culverts_field_names
        )

        add_or_update_datasets(
            line_group,
            culverts_numpy_array_dict,
            culverts_field_names,
            culverts_numpy_array_dict["pk"],
            content_pk,
            ignore_mask=content_type != "v2_culvert",
            field_name_override=culverts_field_name_override,
        )



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

NODE_TYPE__EQ_SUBSETS = {
    "2D_OPEN_WATER": 1,
    "2D_GROUNDWATER": 2,
    "1D_NO_STORAGE": 3,
    "1D_STORAGE": 4,
    "2D_BOUNDARIES": 5,
    "2D_GROUNDWATER_BOUNDARIES": 6,
    "1D_BOUNDARIES": 7,
}

NODE_TYPE__IN_SUBSETS = dict()
NODE_TYPE__IN_SUBSETS["1D"] = [
    NODE_TYPE__EQ_SUBSETS["1D_NO_STORAGE"],
    NODE_TYPE__EQ_SUBSETS["1D_STORAGE"],
]

NODE_TYPE__IN_SUBSETS["1D_ALL"] = [
    NODE_TYPE__EQ_SUBSETS["1D_NO_STORAGE"],
    NODE_TYPE__EQ_SUBSETS["1D_STORAGE"],
    NODE_TYPE__EQ_SUBSETS["1D_BOUNDARIES"],
]

NODE_TYPE__IN_SUBSETS["GROUNDWATER_ALL"] = [
    NODE_TYPE__EQ_SUBSETS["2D_GROUNDWATER"],
    NODE_TYPE__EQ_SUBSETS["2D_GROUNDWATER_BOUNDARIES"],
]

NODE_TYPE__IN_SUBSETS["2D_ALL"] = [
    NODE_TYPE__EQ_SUBSETS["2D_GROUNDWATER"],
    NODE_TYPE__EQ_SUBSETS["2D_GROUNDWATER_BOUNDARIES"],
    NODE_TYPE__EQ_SUBSETS["2D_BOUNDARIES"],
    NODE_TYPE__EQ_SUBSETS["2D_OPEN_WATER"],
]



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
"""
Models
++++++

Breaches are better described as possible breach locations
as they represent the intersection between threedicore
1d-2d flow lines and levee geometries from the models spatialite
database. That is, once the kcu (the type classifier, see :ref:`kcu-label`)
of the 1d-2d flow line has the value 56 there actually will be an active
link between the 1d and the 2d calculation nodes.

To query the ``Breaches`` model::

    >>> from threedigrid.admin.breaches.models import Breaches
    >>> from threedigrid.admin.gridadmin import GridH5Admin

    >>> f = 'gridadmin.h5'
    >>> ga = GridH5Admin(f)
    >>> # get all active breaches
    >>> ga.breaches.filter(kcu__eq=56).data

"""

import logging

from threedigrid.admin import constants
from threedigrid.admin.breaches import exporters
from threedigrid.admin.utils import PKMapper
from threedigrid.orm.base.fields import IndexArrayField
from threedigrid.orm.fields import ArrayField, MultiLineArrayField, PointArrayField
from threedigrid.orm.models import Model

logger = logging.getLogger(__name__)


class LineGeometries(MultiLineArrayField):
    """
    Used for setting `line_geometries` (field in 'lines' h5 group) on the
    Breaches model.
    """

    def __init__(self):
        super().__init__(None)

    def get_value(self, datasource, name, **kwargs):
        if name in list(datasource.keys()):
            return datasource[name]

        if datasource._gridadmin is not None:
            # Return mapped line_coords, based on levl
            levl = datasource["levl"][:]
            data = datasource._gridadmin.lines.filter(id__in=levl).only(
                "id", "line_geometries"
            )
            return PKMapper(data.id, levl).apply_on(data.line_geometries)

        return ArrayField.get_value(datasource, name)


class Breaches(Model):
    """
    fields originating from threedicore:

        - levbr (breach width)
        - levmat (material code)
        - levl (exchange level)
        - kcu (type field 55 or 56)

    added fields from spatialite database:

        - content_pk (primary key database)
        - seq_ids (sequence ids generated during input file generation)

    fields from `lines` h5 group:

        - line_geometries
    """

    content_pk = ArrayField(type=int)
    seq_ids = ArrayField(type=int)
    levbr = ArrayField(type=float)
    levl = IndexArrayField(to="Lines")
    levmat = ArrayField(type=int)
    kcu = ArrayField(type=int)
    line_geometries = LineGeometries()
    coordinates = PointArrayField()
    code = ArrayField(type=str)
    display_name = ArrayField(type=str)

    OBJECT_TYPE = constants.TYPE_V2_BREACH

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._exporters = [
            exporters.BreachesOgrExporter(self),
        ]



#!/usr/bin/env python

"""The setup script."""


import codecs
import os
import re

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

lightweight_requirements = [
    "numpy>=1.15",
    "h5py>=2.9",
    "six",
]

results_requirements = ["cftime>=1.0.1"]

# for extra 'geo'
geo_requirements = [
    "Click>=6.0",
    "mercantile>=1.0.4",
    "pyproj>=2.2",
    "Shapely>=1.6",
    "geojson>=2.4",
]

rpc_requirements = [
    "asyncio-rpc>=0.1.10",
]

setup_requirements = []

test_requirements = ["pytest==3.4.1"]

docs_requirements = [
    "sphinx~=7.2.6",
    "sphinx_rtd_theme>=2.0.0",
]

setup(
    author="Lars Claussen",
    author_email="info@nelen-schuurmans.nl",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
    ],
    description="Python package for the threedigrid administration",
    entry_points={
        "console_scripts": [
            "3digrid_explore=threedigrid.management.commands.kick:kick_start",
            "3digrid_export=threedigrid.management.commands.kick:export_to",
        ],
    },
    install_requires=lightweight_requirements,
    python_requires=">=3.8",
    license="BSD license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="threedigrid",
    name="threedigrid",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    extras_require={
        "geo": geo_requirements,
        "results": results_requirements,
        "rpc": rpc_requirements,
        "docs": docs_requirements,
    },
    url="https://github.com/nens/threedigrid",
    version=find_version("threedigrid", "__init__.py"),
    zip_safe=False,
)



import json
import logging
from collections import OrderedDict

import numpy as np

from threedigrid.admin import constants
from threedigrid.geo_utils import raise_import_exception, transform_bbox
from threedigrid.orm.base.encoder import NumpyEncoder
from threedigrid.orm.base.models import Model

try:
    import geojson
except ImportError:
    geojson = None


logger = logging.getLogger(__name__)


class GeoJsonSerializer:
    def __init__(self, fields, model=None, indent=None):
        if geojson is None:
            raise_import_exception("geojson")
        self.fields = fields
        if model:
            assert isinstance(model, Model)
        self._model = model
        self._indent = indent

    def save(self, filename, **kwargs):
        with open(filename, "w") as file:
            file.write('{"type": "FeatureCollection", "features": [')
            first = True
            for geo in self.geos_iter():
                if not first:
                    file.write(", ")
                else:
                    first = False
                json.dump(
                    geo, file, indent=self._indent, allow_nan=False, cls=NumpyEncoder
                )
            file.write("]}")

    def geos_iter(self):
        if self._model.count == 0:
            return []

        # Get the data, skipping the dummy element
        data = self._model.filter(id__ne=0).to_dict()
        content_type = self._model.__contenttype__()
        model_type = type(self._model).__name__
        if content_type == "lines":
            for i in range(data["id"].shape[-1]):
                linepoints = np.round(
                    data["line_geometries"][i].reshape(2, -1).T.astype("float64"),
                    constants.LONLAT_DIGITS,
                )
                line = geojson.LineString(linepoints.tolist())

                properties = fill_properties(self.fields, data, i, model_type)
                yield geojson.Feature(geometry=line, properties=properties)
        elif content_type in ("nodes", "pumps"):
            for i in range(data["id"].shape[-1]):
                coords = np.round(data["coordinates"][:, i], constants.LONLAT_DIGITS)
                point = geojson.Point([coords[0], coords[1]])
                properties = fill_properties(self.fields, data, i, model_type)
                yield geojson.Feature(geometry=point, properties=properties)
        elif content_type == "breaches":
            for i in range(data["id"].shape[-1]):
                linepoints = np.round(
                    data["line_geometries"][i].reshape(2, -1).T.astype("float64"),
                    constants.LONLAT_DIGITS,
                )
                line = geojson.LineString(linepoints.tolist())
                properties = fill_properties(self.fields, data, i, model_type)
                yield geojson.Feature(geometry=line, properties=properties)
        elif content_type == "cells":
            if (
                self._model.reproject_to_epsg is not None
                and self._model.reproject_to_epsg != self._model.epsg_code
            ):
                cell_coords = transform_bbox(
                    self._model.reproject_to(self._model.epsg_code).cell_coords,
                    self._model.epsg_code,
                    self._model.reproject_to_epsg,
                    all_coords=True,
                )
            else:
                cell_coords = np.array(
                    [
                        data.get("cell_coords")[0],
                        data.get("cell_coords")[3],
                        data.get("cell_coords")[2],
                        data.get("cell_coords")[3],
                        data.get("cell_coords")[2],
                        data.get("cell_coords")[1],
                        data.get("cell_coords")[0],
                        data.get("cell_coords")[1],
                    ]
                )
            cell_coords = np.round(cell_coords, constants.LONLAT_DIGITS)
            for i in range(data["id"].shape[-1]):
                left_top = (cell_coords[0][i], cell_coords[1][i])
                right_top = (cell_coords[2][i], cell_coords[3][i])
                right_bottom = (cell_coords[4][i], cell_coords[5][i])
                left_bottom = (cell_coords[6][i], cell_coords[7][i])
                polygon = geojson.Polygon(
                    [(left_top, right_top, right_bottom, left_bottom, left_top)]
                )
                properties = fill_properties(self.fields, data, i, model_type)
                yield geojson.Feature(geometry=polygon, properties=properties)
        elif content_type == "levees":
            for i in range(data["id"].shape[-1]):
                coords = np.round(
                    data["coords"][i].reshape(2, -1).astype("float64"),
                    constants.LONLAT_DIGITS,
                )
                line = geojson.LineString(coords.T.tolist())
                properties = fill_properties(self.fields, data, i, model_type)
                yield geojson.Feature(geometry=line, properties=properties)
        else:
            raise ValueError("Unknown content type for %s" % self._model)

    @property
    def geos(self):
        return [x for x in self.geos_iter()]

    @property
    def data(self):
        geos = self.geos
        return json.dumps(
            {"type": "FeatureCollection", "features": geos},
            indent=self._indent,
            cls=NumpyEncoder,
        )


def fill_properties(fields, data, index, model_type=None):
    """Returns a dict containing the keys from `fields` filled from `data`
    at `index`

    :param fields: (list) list of keys (str) to extract from data. The keys
        can can also be a dict, in which case fill_properties will be called
        on its items.
    :param data: (dict) the original data
    :param index: (int) index value to extract from data
    :param model_type: (str, optional) optionally adds the given model_type
    """
    result = OrderedDict()
    for field in fields:
        if isinstance(field, dict):
            for key, sub_list in field.items():
                result[key] = fill_properties(sub_list, data, index)
        else:
            field_data = data.get(field, None)
            if field_data is not None and field_data.size > 0:
                if field == "line_geometries" and len(field_data.shape) > 1:
                    # Indexing for reprojected line_geometries
                    value = field_data[index]
                else:
                    value = field_data[..., index]
                # Replace NaN, Inf, -Inf, -9999.0 floats with None (null)
                if np.issubdtype(value.dtype, np.floating):
                    is_invalid = (~np.isfinite(value)) | (value == -9999.0)
                else:
                    is_invalid = False
                if np.any(is_invalid):
                    value = np.where(is_invalid, None, value)
                if value.size == 1:
                    value = value.item()
            else:
                if index == 0:
                    # only log it once
                    logger.warning("missing field %s" % field)
                value = None
            result[field] = value

    if model_type:
        result["model_type"] = model_type

    return result



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

import numpy as np

from threedigrid.geo_utils import (
    get_transformer,
    raise_import_exception,
    select_geoms_by_geometry,
    select_lines_by_tile,
    select_points_by_bbox,
    select_points_by_tile,
    transform_xys,
)
from threedigrid.numpy_utils import (
    angle_in_degrees,
    get_bbox_by_point,
    reshape_flat_array,
    select_lines_by_bbox,
)

from .base.fields import ArrayField

try:
    import shapely
    from shapely.geometry import LineString, Point, Polygon
except ImportError:
    shapely = None


NULL_VALUE = -9999.0


class GeomArrayField(ArrayField):
    """
    Base geometry field
    """

    def reproject(self, values, source_epsg, target_epsg):
        raise NotImplementedError()

    def to_centroid(self, values):
        raise NotImplementedError()

    def get_mask_by_geometry(self, geometry, values):
        geoms = self._to_shapely_geom(values)

        intersects_i = select_geoms_by_geometry(geoms, geometry)

        mask = np.zeros((len(values.T),), dtype=bool)
        mask[np.array(intersects_i, dtype=int)] = True
        return mask

    def _to_shapely_geom(self, values):
        """Returns a list of shapely geometries created from values

        :param values: coordinates
        :return: list of shapely geometries
        """
        raise NotImplementedError()


class PointArrayField(GeomArrayField):
    type = "point"

    def reproject(self, values, source_epsg, target_epsg):
        """
        Reproject the point coordinates: x_array=values[0], y_array=values[1]
        from source_epsg to target_epsg.
        """
        return transform_xys(values[0], values[1], source_epsg, target_epsg)

    def get_mask_by_bbox(self, bbox, values, include_intersections=False):
        """
        Return as boolean mask (np.array) for line/bbox in "values"
            x_array=values[0], y_array=values[1]
        within bbox (x1, y1, x2, y2)
        """
        # For points include_intersections can be ignored
        return select_points_by_bbox(values, bbox)

    def get_mask_by_tile(
        self, tile_xyz, target_epsg, values, include_intersections=False
    ):
        """
        Return as boolean mask (np.array) for points in "values"
        within tile_xyz bbox projected to target_epsg.
        """
        # For points include_intersections can be ignored
        return select_points_by_tile(tile_xyz, target_epsg, values)

    def to_centroid(self, values):
        """
        Returns: the centroid of the point coordinates:
                 x_array=values[0], y_array=values[1]
        """
        return values

    def _to_shapely_geom(self, values):
        if shapely is None:
            raise_import_exception("shapely")

        points = []
        for coord in values.T:
            if np.isnan(coord).all():
                coord = np.full_like(coord, fill_value=NULL_VALUE)

            point = Point(coord[0], coord[1])
            points.append(point)

        return points


class LineArrayField(GeomArrayField):
    type = "line"

    """
    Field which handles line/bbox geoms values:

        x1_array=values[0], y1_array=values[1],
        x2_array=values[2], y2_array=values[3]
    """

    def reproject(self, values, source_epsg, target_epsg):
        """
        Reproject the line/bbox coordinates:
            x1_array=values[0], y1_array=values[1],
            x2_array=values[2], y2_array=values[3]
        from source_epsg to target_epsg.
        """
        return np.vstack(
            (
                transform_xys(values[0], values[1], source_epsg, target_epsg),
                transform_xys(values[2], values[3], source_epsg, target_epsg),
            )
        )

    def get_mask_by_point(self, pnt, values):
        return get_bbox_by_point(pnt, values)

    def get_mask_by_bbox(self, bbox, values, include_intersections=False):
        """
        Return as boolean mask (np.array) for line/bbox in "values"
            x1_array=values[0], y1_array=values[1],
            x2_array=values[2], y2_array=values[3]
        within bbox (x1, y1, x2, y2)
        """
        return select_lines_by_bbox(values, bbox, include_intersections)

    def get_mask_by_tile(
        self, tile_xyz, target_epsg, values, include_intersections=False
    ):
        """
        Return as boolean mask (np.array) for line/bbox in "values"
            x1_array=values[0], y1_array=values[1],
            x2_array=values[2], y2_array=values[3]
        within tile_xyz bbox projected to target_epsg.
        """
        return select_lines_by_tile(
            tile_xyz, target_epsg, values, include_intersections
        )

    def to_centroid(self, values):
        """
        :return: the centroid (float) for the line coordinates:
                    x1_array=values[0], y1_array=values[1],
                    x2_array=values[2], y2_array=values[3]
        """
        return np.array(((values[0] + values[2]) / 2.0, (values[1] + values[3]) / 2.0))

    def get_angles_in_degrees(self, values):
        """
        Returns: the angles in degrees of the lines
        made up by the points:

            x1_array=values[0], y1_array=values[1],
            x2_array=values[2], y2_array=values[3]

        with the (horizontal) x-axis
        """
        if np.isnan(values).all():
            values = np.full_like(values, fill_value=NULL_VALUE)

        return angle_in_degrees(values[0], values[1], values[2], values[3])

    def _to_shapely_geom(self, values):
        if shapely is None:
            raise_import_exception("shapely")

        lines = []
        for coords in values.T:
            if np.isnan(coords).all():
                coords = np.full_like(coords, fill_value=NULL_VALUE)
            line = LineString(coords.reshape((2, -1)))
            lines.append(line)
        return lines


class MultiLineArrayField(GeomArrayField):
    type = "multiline"

    def reproject(self, values, source_epsg, target_epsg):
        """
        Reproject the line/bbox coordinates:
            x1_array=values[0], y1_array=values[1],
            x2_array=values[2], y2_array=values[3]
        from source_epsg to target_epsg.
        """
        reshaped_values = list(map(reshape_flat_array, values))

        transformer = get_transformer(source_epsg, target_epsg)
        transform_values = [
            np.array(transformer.transform(x[0], x[1])).flatten()
            for x in reshaped_values
        ]
        return np.array(transform_values, dtype=object)

    def _to_shapely_geom(self, values):
        if shapely is None:
            raise_import_exception("shapely")

        multilines = []
        for coords in values:
            if np.isnan(coords).all():
                coords = np.full_like(coords, fill_value=NULL_VALUE)

            line = LineString(coords.reshape((2, -1)).T)
            multilines.append(line)
        return multilines


class PolygonArrayField(GeomArrayField):
    """
    Field which handles line/bbox geoms values:

        x1_array=values[0], y1_array=values[1],
        x2_array=values[2], y2_array=values[3]

    """

    type = "polygon"

    def get_mask_by_point(self, pnt, values):
        return get_bbox_by_point(pnt, values)

    def reproject(self, values, source_epsg, target_epsg):
        """
        Reproject the cell coordinates (lower left and
        upper right corners):
            x1_array=values[0], y1_array=values[1],
            x2_array=values[2], y2_array=values[3],
        from source_epsg to target_epsg.
        """
        return np.vstack(
            (
                np.array(transform_xys(values[0], values[1], source_epsg, target_epsg)),
                np.array(transform_xys(values[2], values[3], source_epsg, target_epsg)),
            )
        )

    def _to_shapely_geom(self, values):
        if shapely is None:
            raise_import_exception("shapely")

        polygons = []
        for coords in values:
            if np.isnan(coords).all():
                coords = np.full_like(coords, fill_value=NULL_VALUE)
            polygon = Polygon(coords.reshape((2, -1)).T)
            polygons.append(polygon)
        return polygons


class BboxArrayField(LineArrayField):
    type = "bbox"

    def _to_shapely_geom(self, values):
        if shapely is None:
            raise_import_exception("shapely")

        polygons = []
        for coord in values.T:
            if np.isnan(coord).all():
                coord = np.full_like(coord, fill_value=NULL_VALUE)

            # convert the bbox bottom-left and upper-right into a polygon
            polygon = Polygon(
                [
                    (coord[0], coord[1]),
                    (coord[0], coord[3]),
                    (coord[2], coord[3]),
                    (coord[2], coord[1]),
                ]
            )
            polygons.append(polygon)
        return polygons



"""help text functions for threedigrid management commands"""

from collections import OrderedDict

from threedigrid.admin.gridadmin import GridH5Admin


def model_overview(grid_file):
    """

    :param grid_file: path to hdf5 grid file
    :returns formatted string like so::

        threedicore version:     0-20180219-587b996-1
        has 1d:                  True
        has 2d:                  True
        has groundwater:         False
        has levees:              True
    """

    grid = GridH5Admin(grid_file)
    _attrs = OrderedDict(
        [
            ("model_slug", ""),
            ("threedicore_version", ""),
            ("threedi_version", ""),
            ("has_1d", ""),
            ("has_2d", ""),
            ("has_groundwater", ""),
            ("has_levees", ""),
            ("has_breaches", ""),
            ("has_pumpstations", ""),
        ]
    )

    for k in _attrs.keys():
        if not hasattr(grid, k):
            continue
        attr = getattr(grid, k)
        if attr is None:
            continue
        _attrs[k] = "{:24} {}".format(k.replace("_", " ") + ":", attr)

    return "\n".join(v for v in _attrs.values() if v != "") + "\n"



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.


from threedigrid.orm.base.options import ModelMeta
from threedigrid.orm.base.timeseries_mixin import ResultMixin


class BreachesResultsMixin(ResultMixin):
    class Meta(metaclass=ModelMeta):
        field_attrs = ["units", "long_name", "standard_name"]

        # customize field names
        composite_fields = {
            "breach_depth": ["Mesh1D_breach_depth"],
            "breach_width": ["Mesh1D_breach_width"],
        }

    def __init__(self, **kwargs):
        """Instantiate a breach with netcdf results.

        Variables stored in the netcdf and related to breaches are dynamically
        added as attributes as TimeSeriesArrayField.

        :param kwargs:
        """
        super().__init__(**kwargs)


class BreachesAggregateResultsMixin(ResultMixin):
    class Meta(metaclass=ModelMeta):
        field_attrs = ["units", "long_name"]

        base_composition = {
            "breach_depth": ["Mesh1D_breach_depth"],
            "breach_width": ["Mesh1D_breach_width"],
        }
        composition_vars = {
            "breach_depth": ["avg", "min", "max"],
            "breach_width": ["avg", "min", "max"],
        }

    def __init__(self, **kwargs):
        """Instantiate a line with netcdf results.

        Variables stored in the netcdf and related to lines are dynamically
        added as attributes as TimeSeriesArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super().__init__(**kwargs)
